# -*- coding: utf-8 -*-
"""
Visualization of results using Google Earth


Attributes

category_colours : list
    list of colour codes (aabbggrr) used for nodes and branches. The second
    last value is for NaN, and the last value is default colour. So with
    e.g. 5 colour categories, the list should have 7 elements.

    Colour codes are strings on the format aabbggrr (8-digit hex) - alpha,
    blue, green, red

Example

powergama.GIS.makekml("output.kml",grid_data=data,res=res,
                      nodetype="nodalprice",branchtype="flow")
"""


import simplekml
import math
import numpy
import matplotlib as mpl


category_colours=["ffff6666","ffffff66","ff66ff66","ff66ffff",
            "f6666fff","ffaaaaaa","ff000000"]
'''Default category colours'''

dcbranch_colour = "ffffffff" #white
generator_colour = "ff0000ff" #red
consumer_colour = "ffff00ff" #purple
flowarrow_colour = "ff0000ff" #red

PRICE_MAX = 200
'''default max cap for colouring'''

linewidth = 1.5
'''Default line width'''

point_icon_href = "http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png"
arrow_icon_href = "http://maps.google.com/mapfiles/kml/shapes/donut.png"
'''url of arrow icon'''


def makekml(kmlfile, grid_data,nodetype=None, branchtype=None,
            res=None,timeMaxMin=None,title='PowerGAMA Results'):
    '''Export KML file for Google Earth plot of data

    Colours can be controlled via the module variable "colours"

    Parameters
    ==========
    kmlfile : string
        name of KLM file to create
    grid_data : powergama.GridData
        grid data object
    nodetype : string
        how to plot nodes - 'nodalprice', 'area', 'powergim_type'
    branchtype : string
        how to plot branches -
        'capacity', 'flow', 'sensitivity', 'utilisation',
        'area', 'powergim_type'
    res : powergama.Results (optional)
        result object (result from powergama simulation)
    timeMaxMin : [min,max]
        time range used when plotting results from simulation
    title : string
        title of KML document
    '''
    kml = simplekml.Kml()
    kml.document.name = title

    # Colours, X categories + NaN category + black(default)
    colorbgr_n = category_colours
    colorbgr_b = category_colours
    numCat_n = len(colorbgr_n)-2
    defaultCat_n = numCat_n+1
    numCat_b = len(colorbgr_b)-2
    defaultCat_b = numCat_b+1
    #balloonstyle messes up Google Earth sidebar, for some reason
    #balloontext = "<h3>$[name]</h3> $[description]"

    if nodetype=='area':
        N_categories = len(grid_data.node.area.unique())
        #rgb_colours = [mpl.cm.jet(i/N_categories) for i in range(N_categories)]
        rgb_colours = mpl.cm.tab20(numpy.arange(N_categories)%20)
        colorbgr_n = [simplekml.Color.rgb(int(c[0]*255),int(c[1]*255),int(c[2]*255)) 
                    for c in rgb_colours]
        numCat_n = N_categories
        defaultCat_n=None
    if branchtype=='area':
        N_categories = len(grid_data.node.area.unique())
        #rgb_colours = [mpl.cm.jet(i/N_categories) for i in range(N_categories)]
        rgb_colours = mpl.cm.tab20(numpy.arange(N_categories)%20)
        colorbgr_b = [simplekml.Color.rgb(int(c[0]*255),int(c[1]*255),int(c[2]*255)) 
                    for c in rgb_colours]
        colorbgr_b.append('ffffffff') #white inter-area lines
        numCat_b = N_categories+1
        defaultCat_b=None

    styleNodes = []
    for col in colorbgr_n:
        styleNode = simplekml.Style()
        styleNode.iconstyle.color = col
        styleNode.iconstyle.icon.href = point_icon_href
        styleNode.labelstyle.scale = 0.0 #hide
        #styleNode.balloonstyle.text = balloontext
        styleNodes.append(styleNode)

    styleGenerator = simplekml.Style()
    styleGenerator.iconstyle.icon.href = point_icon_href
    styleGenerator.iconstyle.color  = generator_colour
    styleGenerator.labelstyle.scale = 0.0 #hide
    #styleGenerator.balloonstyle.text = balloontext

    styleConsumer = simplekml.Style()
    styleConsumer.iconstyle.icon.href = point_icon_href
    styleConsumer.iconstyle.color  = consumer_colour
    styleConsumer.labelstyle.scale = 0.0 #hide
    #styleConsumer.iconstyle.visibility = 0 #doesn't work
    #styleConsumer.balloonstyle.text = balloontext

    styleBranches = []
    for col in colorbgr_b:
        styleBranch = simplekml.Style()
        styleBranch.linestyle.color = col
        styleBranch.linestyle.width = linewidth
        styleBranches.append(styleBranch)

    styleDcBranch = simplekml.Style()
    styleDcBranch.linestyle.color = dcbranch_colour
    styleDcBranch.linestyle.width = linewidth

    styleFlowArrow = simplekml.Style()
    styleFlowArrow.iconstyle.icon.href = arrow_icon_href
    styleFlowArrow.iconstyle.color = flowarrow_colour
    styleFlowArrow.iconstyle.scale = 0.5
    styleFlowArrow.labelstyle.scale = 0.0 #hide


    # NODES ##################################################################
    nodefolder = kml.newfolder(name="Node")
    #nodecount = len(grid_data.node.id)
    if nodetype=='nodalprice':
        ## Show nodal price
        meannodalprices = res.getAverageNodalPrices(timeMaxMin)
        # Obs: some values may be numpy.nan
        maxnodalprice = min(PRICE_MAX,numpy.nanmax(meannodalprices))
        minnodalprice = numpy.nanmin(meannodalprices)
        steprange = (maxnodalprice - minnodalprice) / numCat_n
        categoryMax = [math.ceil(minnodalprice+steprange*(n+1))
            for n in range(numCat_n)]
        nodalpricelevelfolder=[]
        for level in range(numCat_n):
            nodalpricelevelfolder.append(nodefolder.newfolder(
                name="Price <= %s" % (str(categoryMax[level]))))
        nodalpricelevelfolder.append(nodefolder.newfolder(
                name="Price NaN" ))
    elif nodetype=='area':
        nodetypes = grid_data.node.area.unique().tolist()
        nodetypefolder=dict()
        for typ in nodetypes:
            nodetypefolder[typ] = nodefolder.newfolder(
                name="Area = {}".format(typ))
    elif nodetype=='powergim_type':
        nodetypes = grid_data.node.type.unique().tolist()
        nodetypefolder=dict()
        for typ in nodetypes:
            nodetypefolder[typ] = nodefolder.newfolder(
                name="Type = {}".format(typ))


    #for i in range(nodecount):
    for i in grid_data.node.index:
        name = grid_data.node.id[i]
        lon = grid_data.node.lon[i]
        lat = grid_data.node.lat[i]
        area = grid_data.node.area[i]
        node_category=defaultCat_n
        #color = None
        description="ID: {} <br/> AREA: {}".format(name,area)
        if nodetype==None:
            pnt = nodefolder.newpoint(name=name,coords=[(lon,lat)],
                                      description=description)
        elif nodetype=='nodalprice':
            nodalprice = meannodalprices[i]
            # Determine category
            node_category=numCat_n
            for category in range(numCat_n):
                if nodalprice <= categoryMax[category]:
                    node_category=category
                    break

            description = """
            Index .. %s             <br/>
            Busname .. %s           <br/>
            Area .. %s              <br/>
            Lon .. %s, Lat .. %s    <br/>
            Price .. %s             <br/>
            """%(str(i),name,area,lon,lat,nodalprice)
            pnt = nodalpricelevelfolder[node_category].newpoint(
                name=name,
                description=description,
                coords=[(lon,lat)])
        elif nodetype=='area':
            typ = grid_data.node.area[i]
            description = """
            Index .. %s             <br/>
            Busname .. %s           <br/>
            Area .. %s              <br/>
            Lon .. %s, Lat .. %s    <br/>
            """%(str(i),name,area,lon,lat)
            node_category = nodetypes.index(typ)
            pnt = nodetypefolder[typ].newpoint(name=name,
                                            description=description,
                                            coords=[(lon,lat)])
        elif nodetype=='powergim_type':
            typ = grid_data.node.type[i]
            description="ID: {}</br>Type: {}".format(name,typ)
            node_category = nodetypes.index(typ)
            pnt = nodetypefolder[typ].newpoint(name=name,
                                            description=description,
                                            coords=[(lon,lat)])
        pnt.style = styleNodes[node_category]
        #pnt.style.iconstyle.color = color
        #pnt.style.iconstyle.icon.href = circle
        #pnt.style.labelstyle.color = "00000000"


    # GENERATORS #############################################################
    genfolder = kml.newfolder(name="Generator")
    gentypeList = grid_data.getAllGeneratorTypes()
    # Create sub-folders according to generator sub-types
    gentypefolders = {typ: genfolder.newfolder(name=typ)
                        for typ in gentypeList}

    # Get generators
    #for i in range(gencount):
    for i in grid_data.generator.index:
        typ = grid_data.generator.type[i]
        cap = grid_data.generator.pmax[i]
        name = "GENERATOR"
        if 'desc' in grid_data.generator.columns:
            name = simplekml.makeunicode.u("GENERATOR {}"
                        .format(grid_data.generator.desc[i]))
        node = grid_data.generator.node[i]
        nodeIndx = grid_data.node.index[grid_data.node.id==node][0]
        lon = grid_data.node.lon[nodeIndx]
        lat = grid_data.node.lat[nodeIndx]
        description = """
            Index .. %s <br/>
            Busname .. %s           <br/>
            Fuel .. %s         <br/>
            Capacity .. %s         <br/>
            Lon .. %s, Lat .. %s    <br/>
            """ % (str(i),node,typ,str(cap),str(lon),str(lat))
        pnt = gentypefolders[typ].newpoint(
            name=name,description=description, coords=[(lon,lat)])
        pnt.style = styleGenerator

    # CONSUMERS #############################################################
    consfolder = kml.newfolder(name="Consumer")

    for i in grid_data.consumer.index:
        avg = grid_data.consumer.demand_avg[i]
        name = "Consumer"
        node = grid_data.consumer.node[i]
        nodeIndx = grid_data.node.index[grid_data.node.id==node][0]
        lon = grid_data.node.lon[nodeIndx]
        lat = grid_data.node.lat[nodeIndx]
        description = """
            Index .. %s <br/>
            Busname .. %s           <br/>
            Avg demand .. %s         <br/>
            Lon .. %s, Lat .. %s    <br/>
            """ % (str(i),node,str(avg),str(lon),str(lat))
        pnt = consfolder.newpoint(
            name=name,description=description, coords=[(lon,lat)])
        pnt.style = styleConsumer


    # BRANCHES ###############################################################
    branchfolder = kml.newfolder(name="Branch")

    if branchtype in ['capacity','flow','utilisation','sensitivity']:
        if res is not None:
            meanflows = res.getAverageBranchFlows(timeMaxMin)
            absbranchflow = meanflows[2]
            brancharrowfolder = branchfolder.newfolder(
                name="Flow direction",visibility=0)
            utilisation = 100*res.getAverageUtilisation(timeMaxMin)
            #sensitiviy is provided for branches with <inf capacity only
            avgsense = numpy.zeros(grid_data.branch.shape[0])
            sens_values = res.getAverageBranchSensitivity(timeMaxMin)
            avgsense[res.idxConstrainedBranchCapacity] = -sens_values
        if branchtype=='flow':
            categoryValue = numpy.asarray(absbranchflow)
            categoryTitle = "Flow"
        elif branchtype=="capacity":
            categoryValue = grid_data.branch['capacity']
            categoryTitle = "Capacity"
        elif branchtype=="utilisation":
            categoryValue = numpy.asarray(utilisation)
            categoryTitle = "Utilisation"
        elif branchtype=='sensitivity':
            categoryValue = numpy.asarray(avgsense)
            categoryTitle = "Sensitivity"

        #Max/min non-infinite value:
        max_value =  max(categoryValue[numpy.isfinite(categoryValue)])
        min_value =  min(categoryValue[numpy.isfinite(categoryValue)])
        steprange = (max_value - min_value) / float(numCat_b)
        categoryMax = [math.ceil(min_value+steprange*(n+1) )
            for n in range(numCat_b)]
        branchlevelfolder=[]
        for level in range(numCat_b):
            branchlevelfolder.append(branchfolder.newfolder(
                name="{} <= {}".format(categoryTitle,
                                       str(categoryMax[level]))))
        branchlevelfolder.append(branchfolder.newfolder(
                name="{} NaN".format(categoryTitle) ))
    elif branchtype=="area":
        branchtypes = grid_data.node.area.unique().tolist()
        branchtypes.append('INTERAREA')
        branchlevelfolder=dict()
        for typ in branchtypes:
            branchlevelfolder[typ] = branchfolder.newfolder(
                name="Area = {}".format(typ))
    elif branchtype=="powergim_type":
        branchtypes = grid_data.branch.type.unique().tolist()
        branchlevelfolder=dict()
        for typ in branchtypes:
            branchlevelfolder[typ] = branchfolder.newfolder(
                name="Type = {}".format(typ))


    for i in grid_data.branch.index:
        startbus = grid_data.branch.node_from[i]
        endbus = grid_data.branch.node_to[i]
        #startbusIndx = (grid_data.node.id == startbus)[0]
        startbusIndx = grid_data.node.index[grid_data.node.id==startbus][0]
        #endbusIndx = (grid_data.node.id == endbus)[0]
        endbusIndx = grid_data.node.index[grid_data.node.id==endbus][0]
        startbuslon = grid_data.node.lon[startbusIndx]
        startbuslat = grid_data.node.lat[startbusIndx]
        endbuslon = grid_data.node.lon[endbusIndx]
        endbuslat = grid_data.node.lat[endbusIndx]
        capacity = grid_data.branch.capacity[i]
        name = "{}=={}".format(startbus,endbus)
        description = "{}<br/>CAPACITY: {}".format(name,capacity)
        branch_category=defaultCat_b

        if branchtype in ['capacity','flow','utilisation','sensitivity']:
            # Determine category
            branch_category=numCat_b
            for category in range(numCat_b):
                if categoryValue[i] <= categoryMax[category]:
                    branch_category=category
                    break

            reactance = grid_data.branch.reactance[i]
            description = """
                Index .. {} <br/>
                Type .. AC <br/>
                Startbus .. {}          <br/>
                Endbus .. {}            <br/>
                Capacity .. {}          <br/>
                Reactance .. {}         <br/>
                """.format(str(i),startbus,endbus,capacity,reactance)
            if res is not None:
                flowAB = meanflows[0][i]
                flowBA = meanflows[1][i]
                description = """{}
                Mean flow .. {:.6g}         <br/>
                Mean flow A to B .. {:.6g}    <br/>
                Mean flow B to A .. {:.6g}    <br/>
                Mean utilisation .. {:.6g} %  <br/>
                Mean sensitivity .. {:.6g}   <br/>
                """.format(description,absbranchflow[i],flowAB,flowBA,
                           utilisation[i],avgsense[i])
            lin = branchlevelfolder[branch_category].newlinestring(name=name,
                  description = description,
                  coords=[(startbuslon,startbuslat),(endbuslon,endbuslat)])

            # Branch flow direction indicator
            if res is not None:
                if (flowAB+flowBA)==0:
                    d = 0.5
                else:
                    d = min(0.9,max(0.1,flowAB/(flowAB+flowBA)))
                arrowcoord = _pointBetween(nodeA=(startbuslon,startbuslat),
                                           nodeB=(endbuslon,endbuslat),
                                           weight=d)
                arrowpoint = brancharrowfolder.newpoint(name=None,
                                                        description=None,
                                                        coords=[arrowcoord],
                                                        visibility=0)
                arrowpoint.style = styleFlowArrow

        elif branchtype=='area':
            msk1 = grid_data.node['id']==grid_data.branch.loc[i,'node_from']
            msk2 = grid_data.node['id']==grid_data.branch.loc[i,'node_to']
            area1 = grid_data.node.loc[msk1,'area'].iloc[0]
            area2 = grid_data.node.loc[msk2,'area'].iloc[0]
            if area1==area2:
                typ = area1
            else:
                typ = "INTERAREA"
            branch_category = branchtypes.index(typ)
            description= """
            {}=={} </br>
            Capacity :   {}
            """.format(startbus,endbus,capacity)
            lin = branchlevelfolder[typ].newlinestring(name=name,
                  description = description,
                  coords=[(startbuslon,startbuslat),(endbuslon,endbuslat)])
        elif branchtype=='powergim_type':
            typ = grid_data.branch.type[i]
            branch_category = branchtypes.index(typ)
            description= """
            {}=={} </br>
            Branch type: {} </br>
            Capacity :   {}
            """.format(startbus,endbus,typ,capacity)
            lin = branchlevelfolder[typ].newlinestring(name=name,
                  description = description,
                  coords=[(startbuslon,startbuslat),(endbuslon,endbuslat)])
        else:
            lin = branchfolder.newlinestring(name=name,
                  description = description,
                  coords=[(startbuslon,startbuslat),(endbuslon,endbuslat)])

        lin.style = styleBranches[branch_category]


    # DC BRANCHES ############################################################
    if grid_data.dcbranch.shape[0] > 0:
        dcbranchfolder = kml.newfolder(name="DC branch")

    for i in grid_data.dcbranch.index:
        startbus = grid_data.dcbranch.node_from[i]
        endbus = grid_data.dcbranch.node_to[i]
        startbusIndx = grid_data.node.index[grid_data.node.id==startbus][0]
        endbusIndx = grid_data.node.index[grid_data.node.id==endbus][0]
        startbuslon = grid_data.node.lon[startbusIndx]
        startbuslat = grid_data.node.lat[startbusIndx]
        endbuslon = grid_data.node.lon[endbusIndx]
        endbuslat = grid_data.node.lat[endbusIndx]
        capacity = grid_data.dcbranch.capacity[i]
        name = "{}=={}".format(startbus,endbus)

        description = """
        Index .. {} <br/>
        Type .. DC <br/>
        Startbus .. {}          <br/>
        Endbus .. {}            <br/>
        Capacity .. {}          <br/>
        """.format(i,startbus,endbus,capacity)
        lin = dcbranchfolder.newlinestring(name=name,
              description = description,
              coords=[(startbuslon,startbuslat),(endbuslon,endbuslat)])

        lin.style = styleDcBranch

    if kmlfile.split(".")[-1]=="kmz":
        kml.savekmz(kmlfile)
    else:
        kml.save(kmlfile)


def _pointBetween(nodeA,nodeB,weight):
    '''computes coords on the line between two points

     [lat lon] = pointBetween(self,lat1,lon1,lat2,lon2,d)

    Parameters
    ----------
        nodeA: dublet (lat,lon)
            latitude/longitude of nodeA (degrees)
        nodeB: dublet (lat,lon)
            latitude/longitude of nodeB (degrees)
        weight: double
            weight=0 is node A, 0.5 is halfway between, and 1 is node B

    Returns
    -------
        (lat,lon): dublet
            coordinates of the point inbetween nodeA and nodeB (degrees)


     ref: http://www.movable-type.co.uk/scripts/latlong.html
    '''
    lat1 = nodeA[0]
    lon1 = nodeA[1]
    lat2 = nodeB[0]
    lon2 = nodeB[1]
    if ((lat1==lat2) and (lon1==lon2)):
        lat = lat1
        lon = lon1
    else:
        #transform to radians
        lat1 = lat1*math.pi/180
        lat2 = lat2*math.pi/180
        lon1 = lon1*math.pi/180
        lon2 = lon2*math.pi/180

        #initial bearing
        y = math.sin(lon2-lon1) * math.cos(lat2)
        x = (math.cos(lat1)*math.sin(lat2)
             - math.sin(lat1)*math.cos(lat2)*math.cos(lon2-lon1) )
        bearing = math.atan2(y, x)

        #angular distance from A to B
        d_tot = (math.acos(math.sin(lat1)*math.sin(lat2)
                 + math.cos(lat1)*math.cos(lat2)*math.cos(lon2-lon1)) )
        d = d_tot*weight

        lat = math.asin(math.sin(lat1)*math.cos(d)
                        +math.cos(lat1)*math.sin(d)*math.cos(bearing) )
        lon = lon1 + math.atan2(math.sin(bearing)*math.sin(d)*math.cos(lat1),
                           math.cos(d)-math.sin(lat1)*math.sin(lat))

        #tansform to degrees
        lat = lat*180/math.pi
        lon = lon*180/math.pi
    return (lat,lon)





# -*- coding: utf-8 -*-
"""
PowerGAMA module containing plotting functions
"""

import folium
import folium.plugins
import branca.colormap
import matplotlib as mpl
import matplotlib.colors
import jinja2
import folium.utilities
import itertools
import pandas as pd
import math


def plotMap(
    pg_data,
    pg_res=None,
    filename=None,
    nodetype=None,
    branchtype=None,
    filter_node=[0, 100],
    filter_branch=None,
    timeMaxMin=None,
    spread_nodes_r=None,
    **kwargs
):
    """
    Plot PowerGAMA data/results on map

    Parameters
    ==========
    pg_data : powergama.GridData
        powergama data object
    pg_res : powergama.Results
        powergama results object
    filename : str
        name of output file (html)
    nodetype : str ('nodalprice','area') or None (default)
        how to colour nodes
    branchtype : str ('utilisation','sensitivity','flow','capacity','type') or None (default)
        how to colour branches
    filter_node : list
        max/min value used for colouring nodes (e.g. nodalprice)
    filter_branch : list
        max/min value used for colouring branches (e.g. utilisation)
    timeMaxMin : [min,max]
        time interval used when showing simulation results
    spread_nodes_r : float (degrees)
        radius (degrees) of circle on which overlapping nodes are
        spread (use eg 0.04)
    kwargs : arguments passed on to folium.Map(...)
    """

    if branca.__version__ <= "0.2.0":
        cmSet1 = branca.colormap.linear.Set1
    else:
        cmSet1 = branca.colormap.linear.Set1_03

    if timeMaxMin is None:
        timeMaxMin = [pg_data.timerange[0], pg_data.timerange[-1] + 1]

    #    nodetype=None,branchtype=None,dcbranchtype=None,
    #                    show_node_labels=False,branch_style='c',latlon=None,
    #                    timeMaxMin=None,
    #                    dotsize=40, filter_node=None, filter_branch=None,
    #                    draw_par_mer=False,showTitle=True, colors=True

    # Add geographic information to branches and generators/consumers
    branch = pg_data.branch.copy()
    dcbranch = pg_data.dcbranch.copy()
    node = pg_data.node.copy()
    generator = pg_data.generator.copy()
    consumer = pg_data.consumer.copy()

    if spread_nodes_r is not None:
        # spread out nodes lying on top of each other
        coords = node[["lat", "lon"]]
        dupl_coords = pd.DataFrame()
        dupl_coords["cumcount"] = coords.groupby(["lat", "lon"]).cumcount()
        dupl_coords["count"] = coords.groupby(["lat", "lon"])["lon"].transform("count")
        for i in node.index:
            n_sum = dupl_coords.loc[i, "count"]
            if n_sum > 1:
                # there are more nodes with the same coordinates
                n = dupl_coords.loc[i, "cumcount"]
                theta = 2 * math.pi / n_sum
                node.loc[i, "lat"] += spread_nodes_r * math.cos(n * theta)
                node.loc[i, "lon"] += spread_nodes_r * math.sin(n * theta)
        # node[['lat','lon']] = coords

    # Careful! Merge may change the order
    branch = branch.merge(
        node[["id", "lat", "lon"]], how="left", left_on="node_from", right_on="id"
    )
    branch = branch.merge(
        node[["id", "lat", "lon"]], how="left", left_on="node_to", right_on="id"
    )
    dcbranch = dcbranch.merge(
        node[["id", "lat", "lon"]], how="left", left_on="node_from", right_on="id"
    )
    dcbranch = dcbranch.merge(
        node[["id", "lat", "lon"]], how="left", left_on="node_to", right_on="id"
    )
    generator = generator.merge(
        node[["id", "lat", "lon"]], how="left", left_on="node", right_on="id"
    )
    consumer = consumer.merge(
        node[["id", "lat", "lon"]], how="left", left_on="node", right_on="id"
    )
    gentypes = pg_data.getAllGeneratorTypes()
    areas = pg_data.getAllAreas()
    node = node.reset_index().rename(columns={"index": "index_orig"})
    node = node.merge(
        pd.DataFrame(areas).reset_index(), how="left", left_on="area", right_on=0
    ).rename(columns={"index": "area_ind"})
    node = node.set_index("index_orig")
    # node.sort_index(inplace=True)
    node["area_ind"] = 0.5 + node["area_ind"] % 10

    if pg_res is not None:
        nodalprices = pg_res.getAverageNodalPrices(timeMaxMin)
        node["nodalprice"] = nodalprices
        branch_sensitivity = pg_res.getAverageBranchSensitivity(timeMaxMin)
        branch_utilisation = pg_res.getAverageUtilisation(timeMaxMin)
        br_ind = pg_data.getIdxBranchesWithFlowConstraints()
        branch["flow"] = pg_res.getAverageBranchFlows(timeMaxMin)[2]
        branch["utilisation"] = branch_utilisation
        branch["sensitivity"] = 0
        branch.loc[branch.index.isin(br_ind), "sensitivity"] = branch_sensitivity
        # DC branch utilisation:
        if dcbranch.shape[0] > 0:
            dcbranch_utilisation = pg_res.getAverageUtilisation(
                timeMaxMin, branchtype="dc"
            )
            dcbranch_sensitivity = pg_res.getAverageBranchSensitivity(
                timeMaxMin, branchtype="dc"
            )
            dcbr_ind = pg_data.getIdxDcBranchesWithFlowConstraints()
            dcbranch["flow"] = pg_res.getAverageBranchFlows(
                timeMaxMin, branchtype="dc"
            )[2]
            dcbranch["utilisation"] = 0
            dcbranch.loc[
                dcbranch.index.isin(dcbr_ind), "utilisation"
            ] = dcbranch_utilisation
            dcbranch["sensitivity"] = 0
            dcbranch.loc[
                dcbranch.index.isin(dcbr_ind), "sensitivity"
            ] = dcbranch_sensitivity

    m = folium.Map(location=[node["lat"].median(), node["lon"].median()], **kwargs)

    callbackNode = """function (row,colour) {
               if (colour=='') {
                   colour=row[3]
               }
               var marker = L.circleMarker(new L.LatLng(row[0],row[1]),
                                           {"radius":3,
                                            "color":colour} );
                      marker.bindPopup(row[2]);
                      return marker;
            }"""
    callbackBranch = """function (row,colour) {
                if (colour=='') {
                    colour=row[3]
                }
                var polyline = L.polyline([row[0],row[1]],
                                          {"color":colour} );
                polyline.bindPopup(row[2]);
                return polyline;
            }"""

    # print("Nodes...")
    if nodetype == "nodalprice":
        value_col = "nodalprice"
        if filter_node is None:
            filter_node = [node[value_col].min(), node[value_col].max()]
        cm_node = branca.colormap.LinearColormap(
            ["green", "yellow", "red"], vmin=filter_node[0], vmax=filter_node[1]
        )
        cm_node.caption = "Nodal price"
        m.add_child(cm_node)
    elif nodetype == "area":
        value_col = "area_ind"
        val_max = node[value_col].max()
        cm_node = cmSet1.scale(0, val_max).to_step(10)
        # cm_node.caption = 'Area'
        # m.add_child(cm_node)
    elif nodetype == "type":
        type_val, types = node["type"].factorize(sort=True)
        node["type_num"] = type_val
        value_col = "type_num"
        val_max = node[value_col].max()
        cm_node = cmSet1.scale(0, val_max).to_step(10)
    else:
        value_col = None

    locationsN = []
    for i, n in node.iterrows():
        if not (n[["lat", "lon"]].isnull().any()):
            data = [n["lat"], n["lon"], "Node={}, area={}".format(n["id"], n["area"])]
            if pg_res is not None:
                data[2] = "{}; nodalprice={:g}".format(data[2], n["nodalprice"])
            if value_col is not None:
                colHex = cm_node(n[value_col])
                data.append(colHex)
                colour = ""
            else:
                colour = "blue"
            locationsN.append(data)
        else:
            print("Missing lat/lon for node index={}".format(i))
    feature_group_Nodes = folium.FeatureGroup(name="Nodes").add_to(m)
    FeatureCollection(
        data=locationsN, callback=callbackNode, addto=feature_group_Nodes, colour=colour
    ).add_to(feature_group_Nodes)

    # print("AC branches...")
    if branchtype == "utilisation":
        value_col = "utilisation"
        if filter_branch is None:
            filter_branch = [0, 1]
        cm_branch = branca.colormap.LinearColormap(
            ["green", "yellow", "red"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch utilisation"
        m.add_child(cm_branch)
    elif branchtype == "sensitivity":
        value_col = "sensitivity"
        if filter_branch is None:
            filter_branch = [branch[value_col].min(), branch[value_col].max()]
        cm_branch = branca.colormap.LinearColormap(
            ["red", "yellow", "green"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch capacity sensitivity"
        m.add_child(cm_branch)
    elif branchtype == "flow":
        value_col = "flow"
        if filter_branch is None:
            filter_branch = [branch[value_col].min(), branch[value_col].max()]
        cm_branch = branca.colormap.LinearColormap(
            ["red", "yellow", "green"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch flow (abs value)"
        m.add_child(cm_branch)
    elif branchtype == "capacity":
        value_col = "capacity"
        if filter_branch is None:
            filter_branch = [branch[value_col].min(), branch[value_col].max()]
        cm_branch = branca.colormap.LinearColormap(
            ["red", "yellow", "green"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch capacity"
        m.add_child(cm_branch)
    elif branchtype == "type":
        type_val, types = branch["type"].factorize(sort=True)
        print(types)
        branch["type_num"] = type_val
        print(branch[["type", "type_num"]])
        value_col = "type_num"
        val_max = branch[value_col].max()
        cm_branch = cmSet1.scale(0, val_max).to_step(10)
    else:
        value_col = None
    locationsB = []
    for i, n in branch.iterrows():
        if (branchtype == "capacity") and (n["capacity"] == 0):
            # skip this branch
            pass
        elif not (n[["lat_x", "lon_x", "lat_y", "lon_y"]].isnull().any()):
            data = [
                [n["lat_x"], n["lon_x"]],
                [n["lat_y"], n["lon_y"]],
                "AC Branch={} ({}-{}), capacity={:g}".format(
                    i, n["node_from"], n["node_to"], n["capacity"]
                ),
            ]
            if branchtype == "type":
                data[2] = "{}; type={}".format(data[2], n["type"])
            if pg_res is not None:
                data[2] = "{}; flow={:g}; utilisation={:g}".format(
                    data[2], n["flow"], n["utilisation"]
                )
            if value_col is not None:
                colHex = cm_branch(n[value_col])
                data.append(colHex)
                colour = ""
            else:
                colour = "black"
            locationsB.append(data)
        else:
            print("Missing lat/lon for node index={}".format(i))
    feature_group_Branches = folium.FeatureGroup(name="AC branches").add_to(m)
    FeatureCollection(
        locationsB, callback=callbackBranch, addto=feature_group_Branches, colour=colour
    ).add_to(feature_group_Branches)

    # print("DC branches...")
    #    if branchtype=="utilisation":
    #        value_col = 'utilisation'
    #        #cm_branch = branca.colormap.LinearColormap(['green', 'yellow', 'red'],
    #        #                                       vmin=0, vmax=1)
    #        #cm_branch.caption = 'AC branch utilisation'
    #        #m.add_child(cm_branch)
    #    else:
    #        value_col=None
    locationsBdc = []
    for i, n in dcbranch.iterrows():
        if not (n[["lat_x", "lon_x", "lat_y", "lon_y"]].isnull().any()):
            data = [
                [n["lat_x"], n["lon_x"]],
                [n["lat_y"], n["lon_y"]],
                "DC Branch={} ({}-{}), capacity={:g}".format(
                    i, n["node_from"], n["node_to"], n["capacity"]
                ),
            ]
            if pg_res is not None:
                data[2] = "{}; flow={:g}; utilisation={:g}".format(
                    data[2], n["flow"], n["utilisation"]
                )
            if value_col is not None:
                colHex = cm_branch(n[value_col])
                data.append(colHex)
                colour = ""
            else:
                colour = "blue"
            locationsBdc.append(data)
        else:
            print("Missing lat/lon for node index={}".format(i))
    feature_group_DcBranches = folium.FeatureGroup(name="DC branches").add_to(m)
    FeatureCollection(
        locationsBdc,
        callback=callbackBranch,
        addto=feature_group_DcBranches,
        colour=colour,
    ).add_to(feature_group_DcBranches)

    # print("Consumers...")
    locationsN = []
    for i, n in consumer.iterrows():
        if not (n[["lat", "lon"]].isnull().any()):
            locationsN.append(
                [
                    n["lat"],
                    n["lon"],
                    "Consumer {} at node={}, avg demand={:g} ({})".format(
                        i, n["node"], n["demand_avg"], n["demand_ref"]
                    ),
                ]
            )
        else:
            print("Missing lat/lon for node index={}".format(i))
    feature_group_Consumer = folium.FeatureGroup(name="Consumers").add_to(m)
    FeatureCollection(
        data=locationsN,
        callback=callbackNode,
        addto=feature_group_Consumer,
        colour="blue",
    ).add_to(feature_group_Consumer)

    # print("Generators...")
    # feature_group_Generator = folium.FeatureGroup(name='Generators').add_to(m)
    ngtypes = max(2, len(gentypes))
    cm_stepG = cmSet1.scale(0, ngtypes - 1).to_step(ngtypes)

    groups = generator.groupby("node")
    feature_group_Generators = folium.FeatureGroup(name="Generators").add_to(m)
    gencluster_icon_create_function = """\
    function(cluster) {
        return L.divIcon({
        html: '<b>' + cluster.getChildCount() + '</b>',
        className: 'marker-cluster marker-cluster-large',
        iconSize: new L.Point(20, 20)
        });
    }"""
    #    for ind,gentype in enumerate(gentypes):
    #        #locationsN=[]
    #        for i,n in generator[generator['type']==gentype].iterrows():
    # Loop by nodes and create marker cluster to avoid generators appearing
    # on top of each other:
    for thenode, genindices in groups.groups.items():
        locationsN = []
        marker_cluster = folium.plugins.MarkerCluster(
            icon_create_function=gencluster_icon_create_function
        )
        marker_cluster.add_to(feature_group_Generators)
        for genind in genindices:
            n = generator.loc[genind]
            gentype = n["type"]
            typeind = gentypes.index(gentype)
            if not (n[["lat", "lon"]].isnull().any()):
                data = [
                    n["lat"],
                    n["lon"],
                    "{}<br>Generator {}: {}, pmax={:g}".format(
                        gentype, genind, n["desc"], n["pmax"]
                    ),
                ]
                col = cm_stepG(typeind)
                data.append(col)
                locationsN.append(data)
            else:
                print("Missing lat/lon for node index={}".format(i))

        # feature_group_GenX = folium.FeatureGroup(name=gentype).add_to(m)
        # FeatureCollection(data=locationsN,callback=callbackNode,
        #                  featuregroup=feature_group_GenX,
        #                  colour="red").add_to(feature_group_GenX)
        FeatureCollection(
            data=locationsN, callback=callbackNode, addto=marker_cluster, colour=""
        ).add_to(marker_cluster)

    legend_generator_html = """
         <div style="position: fixed;
             bottom: 20px; left: 20px;
             border:2px solid grey; z-index:9999; font-size:13px;
             background-color: lightgray">
         &nbsp; <b>Generators</b>"""
    #             bottom: 50px; left: 50px; width: 150px; height: 300px;
    for typeind, gentype in enumerate(gentypes):
        col = cm_stepG(typeind)
        legend_generator_html = (
            '{}<br> &nbsp; <i class="fa fa-circle fa-1x" '
            'style="color:{}">&nbsp;{}</i>'.format(legend_generator_html, col, gentype)
        )
    legend_generator_html = "{}</div>".format(legend_generator_html)
    m.get_root().html.add_child(folium.Element(legend_generator_html))

    folium.LayerControl().add_to(m)

    if filename:
        print("Saving map to file {}".format(filename))
        m.save(filename)

    return m


class FeatureCollection(folium.map.FeatureGroup):
    """
    Add features to a map using in-browser rendering.

    Parameters
    ----------
    data : list
        List of list of shape [[], []]. Data points should be of
        the form [[lat, lng]].
    callback : string, default None
        A string representation of a valid Javascript function
        that will be passed a lat, lon coordinate pair.
    featuregroup : folium.FeatureGroup
        Feature group
    colour : string
        colour
    name : string, default None
        The name of the Layer, as it will appear in LayerControls.
    overlay : bool, default True
        Adds the layer as an optional overlay (True) or the base layer (False).
    control : bool, default True
        Whether the Layer will be included in LayerControls.
    """

    _counts = itertools.count(0)

    def __init__(
        self, data, callback, addto, colour, name=None, overlay=True, control=True
    ):
        super(FeatureCollection, self).__init__(
            name=name, overlay=overlay, control=control
        )
        self._name = "FeatureCollection"
        self._data = data
        self._addto = addto
        self._colour = colour
        self._count = next(self._counts)
        self._callback = "var callback{} = {};".format(self._count, callback)

        self._template = jinja2.Template(
            u"""
            {% macro script(this, kwargs) %}
            {{this._callback}}
            (function(){
                var data = {{this._data}};
                //var map = {{this._parent.get_name()}};
                var addto = {{this._addto.get_name()}};
                var colour = '{{this._colour}}';
                for (var i = 0; i < data.length; i++) {
                    var row = data[i];
                    var feature = callback"""
            + "{}".format(self._count)
            + u"""(row,colour);
                    feature.addTo(addto);
                };
            })();
            {% endmacro %}"""
        )


if __name__ == "__main__":
    plotMap(pg_data, res, filename="mapTest.html")



import pandas as pd
import numpy as np
from scipy.stats.distributions import norm
from sklearn.cluster import KMeans
#from sklearn.metrics.pairwise import pairwise_distances_argmin
#from sklearn.datasets.samples_generator import make_blobs
from sklearn.neighbors import KernelDensity
import sklearn.cluster
import sklearn.preprocessing


def _TMPsample_kmeans(X, samplesize):
    """K-means sampling

    Parameters
    ==========
    X : matrix
        data matrix to sample from
    samplesize : int
        size of sample

    This method relies on sklearn.cluster.KMeans

    """
    """
    TODO: Have to weight the importance, i.e. multiply timeseries with
    installed capacities in order to get proper clustering.
    """

    n_clusters=samplesize
    k_means = KMeans(init='k-means++', n_clusters=n_clusters, n_init=10)
    k_means.fit(X)
    # which cluster nr it belongs to:
    k_means_labels = k_means.labels_
    k_means_cluster_centers = k_means.cluster_centers_
    k_means_labels_unique, X_indecies = np.unique(k_means_labels,
                                                  return_index=True)
    #k_means_predict = k_means.predict(X)

    return k_means_cluster_centers


def _TMPsample_mmatching(X, samplesize):
    """
    The idea is to make e.g. 10000 randomsample-sets of size=samplesize
    from the originial datasat X.

    Choose the sampleset with the lowest objective:
    MINIMIZE [(meanSample - meanX)^2 + (stdvSample - stdvX)^2...]

    in terms of stitistical measures
    """

    return


def _TMPsample_meanshift(X, samplesize):
    """M matching sampling

    Parameters
    ==========
    X : matrix
        data matrix to sample from
    samplesize : int
        size of sample

    This method relies on sklearn.cluster.MeanShift

    It is a centroid-based algorithm, which works by updating candidates
    for centroids to be the mean of the points within a given region.
    These candidates are then filtered in a post-processing stage to
    eliminate near-duplicates to form the final set of centroids.
    """
    #from sklearn.cluster import MeanShift, estimate_bandwidth
    #from sklearn.datasets.samples_generator import make_blobs

    # The following bandwidth can be automatically detected using
    bandwidth = sklearn.cluster.estimate_bandwidth(X, quantile=0.2,
                                                   n_samples=samplesize)

    ms = sklearn.cluster.MeanShift(bandwidth=bandwidth, bin_seeding=True)
    ms.fit(X)
    #labels = ms.labels_
    cluster_centers = ms.cluster_centers_

    #labels_unique = np.unique(labels)
    #n_clusters_ = len(labels_unique)

    #print("number of estimated clusters : %d" % n_clusters_)

    return cluster_centers


def _TMPsample_latinhypercube(X, samplesize):
    """Latin hypercube sampling

    Parameters
    ==========
    X : matrix
        data matrix to sample from
    samplesize : int
        size of sample

    This method relies on pyDOE.lhs(n, [samples, criterion, iterations])

    """
    """
    lhs(n, [samples, criterion, iterations])

    n:an integer that designates the number of factors (required)
    samples: an integer that designates the number of sample points to generate
        for each factor (default: n)
    criterion: a string that tells lhs how to sample the points (default: None,
        which simply randomizes the points within the intervals):
        “center” or “c”: center the points within the sampling intervals
        “maximin” or “m”: maximize the minimum distance between points, but place
                          the point in a randomized location within its interval
        “centermaximin” or “cm”: same as “maximin”, but centered within the intervals
        “correlation” or “corr”: minimize the maximum correlation coefficient
    """
    from pyDOE import lhs
    X_rows = X.shape[0]; X_cols = X.shape[1]
    X_mean = X.mean(); X_std = X.std()
    X_sample = lhs( X_cols , samples=samplesize , criterion='center' )
    kernel=False
    if kernel:
        # Fit data w kernel density
        kde = KernelDensity(kernel='gaussian', bandwidth=0.2).fit(X)
        kde.score_samples(X)
        # random sampling (TODO: fit to latin hypercube sample):
        kde_sample = kde.sample(samplesize)
    else:
        # Fit data w normal distribution
        for i in range(X_cols):
            X_sample[:,i] = norm(loc=X_mean[i] , scale=X_std[i]).ppf(X_sample[:,i])

    return X_sample

def sampleProfileData(data, samplesize, sampling_method):
        """ Sample data from full-year time series

        Parameters
        ==========
        data : GridData object
            Sample from data.profiles
        samplesize : int
            size of sample
        sampling_method : str
            'kmeans', 'uniform',
            EXPERIMENTAL: 'kmeans_scale', 'lhs',  ('mmatching', 'meanshift')


        Returns
        =======
            reduced data matrix according to sample size and method



        """

        """
        Harald:
        TODO: Tidy up - remove irrelevant code.
        -Note: Profiles may also include generator cost profile, not only
        generation and consumption
        -Should we cluster the profiles, or all consumers and generators? This
        is different sine many generators/consumers may use the same profile.
        Using all generators/consumers is more difficult, but probably more
        correct
        -How to determine weight between different types of variations, i.e.
        generation/consumption (MW) vs marginal costs (€)? Using normalised
        profiles with no weighing is one such choice.
        """

        X = data.profiles.copy()

        if sampling_method == 'kmeans':
            """
            Harald preferred method:
            Scale all profiles to have similar variability,
            then cluster, and finally scale back
            """
            # Consider only those profiles which are used in the model:
            profiles_in_use = data.generator['inflow_ref'].append(
                data.generator['fuelcost_ref']).append(
                data.consumer['demand_ref']).unique().tolist()
            X = X[profiles_in_use]

            #scaler = sklearn.preprocessing.MinMaxScaler()
            scaler = sklearn.preprocessing.RobustScaler()
            x_scaled = scaler.fit_transform(X)
            X = pd.DataFrame(data=x_scaled, columns=X.columns,
                          index=X.index)
            km_norm2=sklearn.cluster.KMeans(n_clusters=samplesize,
                                            init='k-means++')
            km_norm2.fit(X)
            km_orig=scaler.inverse_transform(km_norm2.cluster_centers_)
            X_sample = pd.DataFrame(data=km_orig,columns=X.columns)
            return X_sample

        elif sampling_method == 'kmeans_scale':
            print("Using k-means with scaled profiles -> IN PROGRESS")
            #TODO: How to scale prices?
            # Multiply time series for load and VRES with their respective
            # maximum capacities in order to get the correct clusters/samples

            for k in data.profiles.columns.values.tolist():
                ref = k
                pmax = sum(data.generator.pmax[g] for g in range(len(data.generator)) if data.generator.inflow_ref[g] == ref)
                if pmax > 0:
                    X[ref] = data.profiles[ref] * pmax
#            for k,row in self.generator.iterrows():
#                pmax = row['pmax']
#                ref = row['inflow_ref']
#                if X[ref].mean()<1:
#                    X[ref] = self.profiles[ref] * pmax
            X['const'] = 1

            for k,row in data.consumer.iterrows():
                pmax = row['demand_avg']
                ref = row['demand_ref']
                X[ref] = data.profiles[ref] * pmax

                X_sample = _TMPsample_kmeans(X, samplesize)
                X_sample = pd.DataFrame(data=X_sample,
                        columns=X.columns)

            # convert back to relative values
            for k in data.profiles.columns.values.tolist():
                ref = k
                pmax = sum(data.generator.pmax[g] for g in range(len(data.generator)) if data.generator.inflow_ref[g] == ref)
                if pmax > 0:
                    X_sample[ref] = X_sample[ref] / pmax
#            for k,row in self.generator.iterrows():
#                pmax = row['pmax']
#                ref = row['inflow_ref']
#                if pmax == 0:
#                    centroids[ref] = 0
#                else:
#                    if X[ref].mean()>1:
#                        centroids[ref] = centroids[ref] / pmax
            for k,row in data.consumer.iterrows():
                pmax = row['demand_avg']
                ref = row['demand_ref']
                if pmax == 0:
                    X_sample[ref] = 0
                else:
                    X_sample[ref] = X_sample[ref] / pmax
            X_sample['const'] = 1
            return X_sample

        elif sampling_method == 'mmatching':
            print("Using moment matching... -> NOT IMPLEMENTED")
        elif sampling_method == 'meanshift':
            print("Using Mean-Shift... -> EXPERIMENTAL")
            X_sample = _TMPsample_meanshift(X, samplesize)
            return X_sample
        elif sampling_method == 'lhs':
            print("Using Latin-Hypercube... -> EXPERIMENTAL")
            X_sample = _TMPsample_latinhypercube(X, samplesize)
            X_sample = pd.DataFrame(data=X_sample,
                        columns=X.columns)
            X_sample['const'] = 1
            X_sample[(X_sample < 0)] = 0
            return X_sample
        elif sampling_method == 'uniform':
            print("Using uniform sampling (consider changing sampling method!)...")
            #Use numpy random in order to have control of ranom seed from
            # top level script (np.random.seed(..))
            timerange=np.random.choice(data.profiles.shape[0],
                                        size=samplesize,replace=False)

            #timerange = random.sample(range(8760),samplesize)
            X_sample = data.profiles.loc[timerange, :]
            X_sample.index = list(range(len(X_sample.index)))
            return X_sample
        else:
            raise Exception("Unknown sampling method")
        return



# -*- coding: utf-8 -*-
'''
Module containing print functions
'''

#import matplotlib.pyplot as plt
#import matplotlib.cm as cm
#import matplotlib as mpl
#import numpy as np

import math

#import powergama.database as db
#import csv
    
def PrintTime(text,time,indent=1):
    
    HoursUsed = math.floor(time/60/60)
    MinutesUsed = math.floor(((time/60/60)-HoursUsed)*60)
    SecondsUsed = int(round(((((time/60/60)-HoursUsed)*60)-MinutesUsed)*60))
    
    if indent==0:
        indentstring=""
    elif indent==1:
        indentstring="    "
    elif indent==2:
        indentstring="        "
    elif indent==3:
        indentstring="            "
    elif indent==4:
        indentstring="                "
    else:
        print("WRONG INDENT SPECIFICATION")
        
    if HoursUsed==0:
        if MinutesUsed==0:
            print("\n{}{}{} seconds\n"
                  .format(indentstring,text, SecondsUsed))
        else:
            print("\n{}{}{} minutes and {} seconds\n"
                  .format(indentstring,text, MinutesUsed, SecondsUsed))
    else:
        print("\n{}{}{} hours, {} minutes and {} seconds\n"
                  .format(indentstring,text, HoursUsed, MinutesUsed, SecondsUsed))
  
    
 

def Print(text,indent=1):
      
    if indent==0:
        indentstring=""
    elif indent==1:
        indentstring="    "
    elif indent==2:
        indentstring="        "
    elif indent==3:
        indentstring="            "
    elif indent==4:
        indentstring="                "
    else:
        print("WRONG INDENT SPECIFICATION")
        

    print("\n{}{}\n".format(indentstring,text))


























# -*- coding: utf-8 -*-
'''
Module containing PowerGAMA LpProblem class
'''

'''
 Power flow equations:

 Linearised ("DC") power flow equation
 Pinj - Bprime * theta = 0
           Bprime = (N-1)x(N-1) matrix (removed ref.bus row/column)
           theta = phase angles (at N-1 buses)
           Pinj = generation - load at node (cf makeSbus)

 Relationship between angles and power flow
 Pb - (D x A) x theta = 0
           theta_j = phase angle node j (excluding ref. node)
           Pb_k = power flow branch k
           D = diag(-b_k) (negative of susceptance on branch k)
           A = Mx(N-1) node-branch incidence (adjacency) matrix
'''

import pyomo.environ as pyo
import pyomo.opt
import numpy as np
#from numpy import pi, array, asarray, vstack, zeros
import datetime
from . import constants as const
import scipy.sparse
import sys
import warnings
import pandas as pd
import networkx as nx
import powergama

#needed for code to work both for python 2.7 and 3:
try:
    from itertools import izip as zip
except ImportError: # will be 3.x series
    pass



class LpProblem(object):
    '''
    Class containing problem definition as a LP problem, and function calls
    to solve the problem

    '''

    def _createAbstractModel(self):
        model = pyo.AbstractModel()
        model.name = 'PowerGAMA abstract model {}'.format(
            datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S"))

        # SETS ###############################################################
        model.NODE = pyo.Set(ordered=True)
        model.GEN = pyo.Set(ordered=True)
        model.GEN_PUMP = pyo.Set(ordered=True)
        model.BRANCH_AC = pyo.Set(ordered=True)
        model.BRANCH_DC = pyo.Set(ordered=True)
        model.LOAD = pyo.Set(ordered=True)
        model.LOAD_FLEX = pyo.Set(ordered=True)

        # PARAMETERS #########################################################
        model.genCost = pyo.Param(model.GEN, within=pyo.Reals,
                                  mutable=True)
        model.pumpCost = pyo.Param(model.GEN_PUMP, within=pyo.Reals,
                                  mutable=True)
        model.flexLoadCost = pyo.Param(model.LOAD_FLEX,
                                       within=pyo.Reals,
                                       mutable=True)
        model.loadShedCost = pyo.Param(within=pyo.NonNegativeReals)
        model.branchAcCapacity = pyo.Param(model.BRANCH_AC,
                                         within=pyo.NonNegativeReals)
        model.branchDcCapacity = pyo.Param(model.BRANCH_DC,
                                         within=pyo.NonNegativeReals)
        model.genPmaxLimit = pyo.Param(model.GEN,within=pyo.NonNegativeReals,
                                       mutable = True)
        model.genPminLimit = pyo.Param(model.GEN,within=pyo.NonNegativeReals,
                                       mutable = True)
        model.pumpCapacity = pyo.Param(model.GEN_PUMP,within=pyo.Reals)
        model.flexloadMax = pyo.Param(model.LOAD_FLEX,
                                       within=pyo.NonNegativeReals)
        model.refNodes = pyo.Param(model.NODE, within=pyo.Boolean)

        # helpers:
        model.genNode = pyo.Param(model.GEN,within=model.NODE)
        model.demNode = pyo.Param(model.LOAD,within=model.NODE)
        model.branchNodeFrom = pyo.Param(model.BRANCH_AC,within=model.NODE)
        model.branchNodeTo = pyo.Param(model.BRANCH_AC,within=model.NODE)
        model.dcbranchNodeFrom = pyo.Param(model.BRANCH_AC,within=model.NODE)
        model.dcbranchNodeTo = pyo.Param(model.BRANCH_AC,within=model.NODE)
        model.demand = pyo.Param(model.LOAD,within=pyo.Reals,
                                 mutable=True)
        model.coeff_B = pyo.Param(model.NODE,model.NODE,within=pyo.Reals)
        model.coeff_DA = pyo.Param(model.BRANCH_AC,model.NODE,within=pyo.Reals)

        if self._lossmethod==1:
            model.lossAcA = pyo.Param(model.BRANCH_AC,within=pyo.Reals,
                                        default=0,mutable=False)
            model.lossAcB = pyo.Param(model.BRANCH_AC,within=pyo.Reals,
                                        default=0,mutable=False)
            model.lossDcA = pyo.Param(model.BRANCH_DC,within=pyo.Reals,
                                        default=0,mutable=False)
            model.lossDcB = pyo.Param(model.BRANCH_DC,within=pyo.Reals,
                                        default=0,mutable=False)
        elif self._lossmethod==2:
            model.branchAcPowerLoss = pyo.Param(model.BRANCH_AC,within=pyo.Reals,
                                                default=0, mutable=True)
            model.branchDcPowerLoss = pyo.Param(model.BRANCH_AC,within=pyo.Reals,
                                                default=0, mutable=True)
            
        # VARIABLES ##########################################################
        model.varAcBranchFlow = pyo.Var(model.BRANCH_AC,within = pyo.Reals)
        model.varDcBranchFlow = pyo.Var(model.BRANCH_AC,within = pyo.Reals)
        if self._lossmethod==1:
            model.varAcBranchFlow12 = pyo.Var(model.BRANCH_AC,
                                              within = pyo.NonNegativeReals)
            model.varAcBranchFlow21 = pyo.Var(model.BRANCH_AC,
                                              within = pyo.NonNegativeReals)
            model.varDcBranchFlow12 = pyo.Var(model.BRANCH_DC,
                                              within = pyo.NonNegativeReals)
            model.varDcBranchFlow21 = pyo.Var(model.BRANCH_DC,
                                              within = pyo.NonNegativeReals)
            model.varLossAc12 = pyo.Var(model.BRANCH_AC,
                                              within = pyo.NonNegativeReals)
            model.varLossAc21 = pyo.Var(model.BRANCH_AC,
                                              within = pyo.NonNegativeReals)
            model.varLossDc12 = pyo.Var(model.BRANCH_DC,
                                              within = pyo.NonNegativeReals)
            model.varLossDc21 = pyo.Var(model.BRANCH_DC,
                                              within = pyo.NonNegativeReals)
        model.varGeneration = pyo.Var(model.GEN,within = pyo.NonNegativeReals)
        model.varPump = pyo.Var(model.GEN_PUMP, within = pyo.NonNegativeReals)
        model.varCurtailment  = pyo.Var(model.GEN,
                                        within = pyo.NonNegativeReals)
        model.varFlexLoad = pyo.Var(model.LOAD_FLEX,
                                    within = pyo.NonNegativeReals)
        model.varLoadShed = pyo.Var(model.LOAD, within = pyo.NonNegativeReals)
        model.varVoltageAngle = pyo.Var(model.NODE, within = pyo.Reals,
                                        initialize=0.0)
# I wonder if these bound on voltage angle creates infeasibility
# - is it really needed
# TODO: Verify voltage angle bounds required
#                                        bounds = (-np.pi,np.pi))


        # CONSTRAINTS ########################################################

        # 1 Power flow limit
        def maxflowAc_rule(model, j):
            cap = model.branchAcCapacity[j]
            if  not np.isinf(cap):
                expr = pyo.inequality(-cap, model.varAcBranchFlow[j], cap )
            else:
                expr = pyo.Constraint.Skip
            return expr
        def maxflowDc_rule(model, j):
            cap = model.branchDcCapacity[j]
            if  not np.isinf(cap):
                expr = pyo.inequality(-cap, model.varDcBranchFlow[j], cap )
            else:
                expr = pyo.Constraint.Skip
            return expr
        model.cMaxFlowAc = pyo.Constraint(model.BRANCH_AC, rule=maxflowAc_rule)
        model.cMaxFlowDc = pyo.Constraint(model.BRANCH_DC, rule=maxflowDc_rule)

        if self._lossmethod==1:
            def flowAc_rule(model, j):
                expr = (model.varAcBranchFlow[j] == 
                        model.varAcBranchFlow12[j]- model.varAcBranchFlow21[j])
                return expr
            def flowDc_rule(model, j):
                expr = (model.varDcBranchFlow[j] == 
                        model.varDcBranchFlow12[j] - model.varDcBranchFlow21[j])
                return expr
            model.cFlowAc = pyo.Constraint(model.BRANCH_AC, rule=flowAc_rule)
            model.cFlowDc = pyo.Constraint(model.BRANCH_DC, rule=flowDc_rule)

        # 1b Losses vs flow
        if self._lossmethod==1:
            def lossAc_rule12(model,j):
                expr = (model.varLossAc12[j] == 
                        model.varAcBranchFlow12[j] * model.lossAcA[j] 
                        + model.lossAcB[j])
                return expr
            def lossAc_rule21(model,j):
                expr = (model.varLossAc21[j] == 
                        model.varAcBranchFlow21[j] * model.lossAcA[j] 
                        + model.lossAcB[j])
                return expr
            def lossDc_rule12(model,j):
                expr = (model.varLossDc12[j] == 
                        model.varDcBranchFlow12[j] * model.lossDcA[j] 
                        + model.lossDcB[j])
                return expr
            def lossDc_rule21(model,j):
                expr = (model.varLossDc21[j] == 
                        model.varDcBranchFlow21[j] * model.lossDcA[j] 
                        + model.lossDcB[j])
                return expr
            model.cLossAc12 = pyo.Constraint(model.BRANCH_AC, rule=lossAc_rule12)
            model.cLossAc21 = pyo.Constraint(model.BRANCH_AC, rule=lossAc_rule21)    
            model.cLossDc12 = pyo.Constraint(model.BRANCH_DC, rule=lossDc_rule12)
            model.cLossDc21 = pyo.Constraint(model.BRANCH_DC, rule=lossDc_rule21)


        # 2 Generator output limit
        # Generator output constraint is not necessary, as lower and upper
        # bounds are set for each timestep in _update_progress. Should not
        # be specified as constraint with with pmax as limit, since e.g.
        # PV may have higher production than generator rating.
        
        #HGS: Doing it anyway, cf Espen Bødal and Martin Kristiansen
        #TODO: Check that there are no problems with this.
        def genMaxLimit_rule(model, i):
            return model.varGeneration[i] <= model.genPmaxLimit[i]
        model.cGenMaxLimit = pyo.Constraint(model.GEN, rule=genMaxLimit_rule)

        def genMinLimit_rule(model, i):
            if model.genPminLimit[i].value > 0:
                return model.varGeneration[i] >= model.genPminLimit[i]
            else:
                return pyo.Constraint.Skip
        model.cGenMinLimit = pyo.Constraint(model.GEN, rule=genMinLimit_rule)

        # 3 Pump output limit
        def pump_rule(model,g):
            expr = (model.varPump[g] <= model.pumpCapacity[g])
            return expr

        model.cPump = pyo.Constraint(model.GEN_PUMP, rule=pump_rule)

        # 4 Flexible load limit
        def flexload_rule(model,i):
            expr = (model.varFlexLoad[i] <= model.flexloadMax[i])
            return expr

        model.cFlexload = pyo.Constraint(model.LOAD_FLEX,rule=flexload_rule)

        #5 Power balance (power flow equation)  (Pnode = B theta)
        #TODO: Speed this up by not looping over all generators etc for each
        # node, use instead a dictionary with key=node, value=list of generators
        # make it from grp=pg_data.generator.groupby('node')
        def powerbalance_rule(model,n):
            lhs = 0
            for g in model.GEN:
                if model.genNode[g]==n:
                    lhs += model.varGeneration[g]
            for p in model.GEN_PUMP:
                if model.genNode[p]==n:
                    lhs -= model.varPump[p]
            for l in model.LOAD:
                if model.demNode[l]==n:
                    lhs -= model.demand[l]
                    lhs += model.varLoadShed[l]
            for f in model.LOAD_FLEX:
                if model.demNode[f]==n:
                    lhs -= model.varFlexLoad[f]
            for b in model.BRANCH_DC:
                # positive sign for flow into node
                if model.dcbranchNodeTo[b]==n:
                    lhs += model.varDcBranchFlow[b]
                    if self._lossmethod==1:
                        lhs -= model.varLossDc12[b]
                    elif self._lossmethod==2:
                        lhs -= model.branchDcPowerLoss[b]/2
                elif model.dcbranchNodeFrom[b]==n:
                    lhs += -model.varDcBranchFlow[b] 
                    if self._lossmethod==1:
                        lhs -= model.varLossDc21[b]
                    elif self._lossmethod==2:
                        lhs -= model.branchDcPowerLoss[b]/2
            if self._lossmethod==1:
                for b in model.BRANCH_AC:
                    # positive sign for flow into node
                    if model.branchNodeTo[b]==n:
                        lhs += -model.varLossAc12[b]
                    elif model.branchNodeFrom[b]==n:
                        lhs += -model.varLossAc21[b]
            elif self._lossmethod==2:
                for b in model.BRANCH_AC:
                    # positive sign for flow into node
                    if model.branchNodeTo[b]==n:
                        lhs -= model.branchAcPowerLoss[b]/2
                    elif model.branchNodeFrom[b]==n:
                        lhs -= model.branchAcPowerLoss[b]/2

            lhs = lhs/const.baseMVA

            rhs = 0
            n2s = [k[1]  for k in model.coeff_B.keys() if k[0]==n]
            for n2 in n2s:
                rhs -= model.coeff_B[n,n2]*(
                            model.varVoltageAngle[n2]*const.baseAngle)
            expr = (lhs == rhs)
            #Skip constraint if it is trivial (otherwise run-time error)
            #TODO: Check if this is safe
            if ((type(expr) is bool) and (expr==True)):
                expr = pyo.Constraint.Skip
            return expr

        model.cPowerbalance = pyo.Constraint(model.NODE,rule=powerbalance_rule)

        #6 Power balance (power flow vs voltage angle)
        def flowangle_rule(model,b):
            lhs = model.varAcBranchFlow[b]
            lhs = lhs/const.baseMVA
            rhs = 0
            #TODO speed up- remove for loop
            n2s = [k[1]  for k in model.coeff_DA.keys() if k[0]==b]
            for n2 in n2s:
                rhs += model.coeff_DA[b,n2]*(
                            model.varVoltageAngle[n2]*const.baseAngle)
            #for n2 in model.NODE:
            #    if (b,n2) in model.coeff_DA.keys():
            #        rhs += model.coeff_DA[b,n2]*model.varVoltageAngle[n2]
            expr = (lhs==rhs)
            return expr

        model.cFlowAngle = pyo.Constraint(model.BRANCH_AC, rule=flowangle_rule)

        #7 Reference voltag angle)
        def referenceNode_rule(model,n):
            if n in model.refNodes.keys():
                expr = (model.varVoltageAngle[n] == 0)
            else:
                expr = pyo.Constraint.Skip
            return expr

        model.cReferenceNode = pyo.Constraint(model.NODE,
                                              rule=referenceNode_rule)

        # OBJECTIVE ##########################################################

        def cost_rule(model):
            """Operational costs: cost of gen, load shed and curtailment"""

            # Operational costs phase 1 (if stage2DeltaTime>0)
            cost = sum(model.varGeneration[i]*model.genCost[i]
                            for i in model.GEN)
            cost -= sum(model.varPump[i]*model.pumpCost[i]
                            for i in model.GEN_PUMP)
            cost -= sum(model.varFlexLoad[i]*model.flexLoadCost[i]
                            for i in model.LOAD_FLEX )
            cost += sum(model.varLoadShed[i]*model.loadShedCost
                            for i in model.LOAD )
            #cost += sum(model.varCurtailment[i,t]*model.curtailmentCost
            #            for i in model.GEN for t in model.TIME)

            return cost

        model.OBJ = pyo.Objective(rule=cost_rule, sense=pyo.minimize)


        return model



    def _createModelData(self,grid_data):
        '''Create model data in dictionary format

        Parameters
        ----------
        grid_data : powergama.GridData object
            contains grid model

        Returns
        --------
        dictionary with pyomo data (in pyomo format)
        '''

        #to see how the data format is:
        #data = pyo.DataPortal(model=self.abstractmodel)
        #data.load(filename=datafile)

        di = {}
        #Sets:
        di['NODE'] = {None: grid_data.node['id'].tolist() }
        di['BRANCH_AC'] = {None: grid_data.branch.index.tolist() }
        di['BRANCH_DC'] = {None: grid_data.dcbranch.index.tolist() }
        di['GEN'] = {None: grid_data.generator.index.tolist() }
        di['GEN_PUMP'] = {None: grid_data.getIdxGeneratorsWithPumping() }
        di['LOAD'] = {None: grid_data.consumer.index.tolist() }
        di['LOAD_FLEX'] = {None: grid_data.getIdxConsumersWithFlexibleLoad() }
        di['AREA'] = {None: grid_data.getAllAreas() }

        # PARAMETERS #########################################################

        #self._marginalcosts_flexload = asarray(grid.consumer['flex_basevalue'])
        di['branchAcCapacity'] = grid_data.branch['capacity'].to_dict()
        di['branchDcCapacity'] = grid_data.dcbranch['capacity'].to_dict()
        di['genPmaxLimit'] = grid_data.generator['pmax'].to_dict()
        di['genPminLimit'] = grid_data.generator['pmin'].to_dict()
        di['genCost'] = grid_data.generator['fuelcost'].to_dict()
        di['pumpCost'] = {k: grid_data.generator['fuelcost'][k]
                            for k in di['GEN_PUMP'][None] }
        di['pumpCapacity'] = {k: grid_data.generator['pump_cap'][k]
                                for k in di['GEN_PUMP'][None] }
        di['flexLoadCost'] = {i: grid_data.consumer['flex_basevalue'][i]
                                for i in di['LOAD_FLEX'][None] }
        di['flexloadMax'] = {i: (grid_data.consumer['demand_avg'][i]
                                    * grid_data.consumer['flex_fraction'][i]
                                    / grid_data.consumer['flex_on_off'][i])
                                for i in di['LOAD_FLEX'][None] }
        # use fixed load shedding penalty of 1000 €/MWh
        di['loadShedCost'] = {None: powergama.constants.loadshedcost}

        di['genNode'] = grid_data.generator['node'].to_dict()
        di['demNode'] = grid_data.consumer['node'].to_dict()
        di['branchNodeFrom'] = grid_data.branch['node_from'].to_dict()
        di['branchNodeTo'] = grid_data.branch['node_to'].to_dict()
        di['dcbranchNodeFrom'] = grid_data.dcbranch['node_from'].to_dict()
        di['dcbranchNodeTo'] = grid_data.dcbranch['node_to'].to_dict()
        di['demand'] = grid_data.consumer['demand_avg'].to_dict()

        if self._lossmethod==1:
            # Upper capacity limit, since capacity may be infinit
            clip_mw = 500
            br = grid_data.branch
            di['lossAcA'] = (
                br['resistance']*br['capacity'].clip(upper=clip_mw)
                /const.baseMVA).to_dict()
            di['lossAcB'] = {b: 0 for b in di['BRANCH_AC'][None] }

            br = grid_data.dcbranch
            di['lossDcA'] = ( 
                br['resistance']*br['capacity'].clip(upper=clip_mw)
                /const.baseMVA).to_dict()
            di['lossDcB'] = {b: 0 for b in di['BRANCH_DC'][None] }

        # Compute matrices used in power flow equaions
        print("Computing B and DA matrices...")
        Bbus, DA = grid_data.compute_power_flow_matrices()

        n_i = di['NODE'][None]
        b_i = di['BRANCH_AC'][None]
        di['coeff_B'] = dict()
        di['coeff_DA'] = dict()

        print("Creating B and DA coefficients...")
        cx = scipy.sparse.coo_matrix(Bbus)
        for i,j,v in zip(cx.row, cx.col, cx.data):
            di['coeff_B'][(n_i[i],n_i[j])] = v

        cx = scipy.sparse.coo_matrix(DA)
        for i,j,v in zip(cx.row, cx.col, cx.data):
            di['coeff_DA'][(b_i[i],n_i[j])] = v

        # Find synchronous areas and specify reference node in each area
        G = nx.Graph()
        G.add_nodes_from(grid_data.node['id'])
        G.add_edges_from(zip(grid_data.branch['node_from'],
                             grid_data.branch['node_to']))

        G_subs = (G.subgraph(c) for c in nx.connected_components(G))
        #deprecated:
        #G_subs = nx.connected_component_subgraphs(G)
        refnodes = []
        for gr in G_subs:
            refnode = list(gr.nodes)[0]
            refnodes.append(refnode)
            print("Found synchronous area (size = {}), using ref node = {}"
                    .format(gr.order(),refnode))
        # use first node as voltage angle reference
        di['refNodes'] = {n:True for n in refnodes}

        return {'powergama':di}


    def __init__(self,grid,lossmethod=0):
        '''LP problem formulation

        Parameters
        ==========
        grid : GridData
            grid data object
        lossmethod : int
            loss method; 0=no losses, 1=linearised losses, 2=added as load
        '''
        self._lossmethod = lossmethod

        # Pyomo
        # 1 create abstract pyomo model
        self.abstractmodel = self._createAbstractModel()

        # 2 create concrete instance using grid data
        modeldata_dict = self._createModelData(grid)
        print('Creating LP problem instance...')
        self.concretemodel = self.abstractmodel.create_instance(
                                data=modeldata_dict,
                                name="PowerGAMA Model",
                                namespace='powergama')

        print('Initialising LP problem...')

        # Creating local variables to keep track of storage
        self._grid = grid
        self.timeDelta = grid.timeDelta
        self._idx_generatorsWithPumping = grid.getIdxGeneratorsWithPumping()
        self._idx_generatorsWithStorage = grid.getIdxGeneratorsWithStorage()

        self._idx_consumersWithFlexLoad = (
            grid.getIdxConsumersWithFlexibleLoad() )
        self._idx_branchesWithConstraints = (
            grid.getIdxBranchesWithFlowConstraints() )
        self._fancy_progressbar = False

        # Initial values of marginal costs, storage and storage values
        self._storage = (
            grid.generator['storage_ini']*grid.generator['storage_cap'] 
            ).fillna(0)

        self._storage_flexload = (
                grid.consumer['flex_storagelevel_init']
                * grid.consumer['flex_storage']
                * grid.consumer['flex_fraction']
                * grid.consumer['demand_avg']
                ).fillna(0)

        self._energyspilled = grid.generator['storage_cap'].copy(deep=True)
        self._energyspilled[:]=0

        return


    def _updateLpProblem(self,timestep):
        '''
        Function that updates LP problem for a given timestep, due to changed
        power demand, power inflow and marginal generator costs
        '''


        # 1. Update bounds on maximum and minimum production (power inflow)
        P_storage = self._storage / self.timeDelta
        P_max = self._grid.generator['pmax']
        P_min = self._grid.generator['pmin']
        for i in self.concretemodel.GEN:
            inflow_factor = self._grid.generator['inflow_fac'][i]
            capacity = self._grid.generator['pmax'][i]
            inflow_profile = self._grid.generator['inflow_ref'][i]
            P_inflow =  (capacity * inflow_factor
                * self._grid.profiles[inflow_profile][timestep])
            if i not in self._idx_generatorsWithStorage:
                '''
                Don't let P_max limit the output (e.g. solar PV)
                This won't affect fuel based generators with zero storage,
                since these should have inflow=p_max in any case
                '''
                if P_min[i] > 0:
                    self.concretemodel.genPminLimit[i] = min(
                        P_inflow,P_min[i]) # Espen
                self.concretemodel.genPmaxLimit[i] = P_inflow # Espen
            else:
                #generator has storage
                if P_min[i] > 0:
                    self.concretemodel.genPminLimit[i] = min(
                        max(0,P_inflow+P_storage[i]), P_min[i]) # Espen
                self.concretemodel.genPmaxLimit[i] = min(
                    max(0,P_inflow+P_storage[i]), P_max[i]) # Espen


        # 2. Update demand (which affects powr balance constraint)
        for i in self.concretemodel.LOAD:
            #dem = self.concretemodel.demand[i]
            average = self._grid.consumer['demand_avg'][i]*(
                            1-self._grid.consumer['flex_fraction'][i])
            profile_ref = self._grid.consumer['demand_ref'][i]
            dem_new = self._grid.profiles[profile_ref][timestep] * average
            self.concretemodel.demand[i] = dem_new

        # 3. Update cost parameters (which affect the objective function)
        # 3a. generators with storage (storage value)
        for i in self._idx_generatorsWithStorage:
            this_type_filling = self._grid.generator['storval_filling_ref'][i]
            this_type_time = self._grid.generator['storval_time_ref'][i]
            storagecapacity = self._grid.generator['storage_cap'][i]
            fillinglevel = self._storage[i] / storagecapacity
            filling_col = int(round(fillinglevel*100))
            storagevalue = (
                self._grid.generator['storage_price'][i]
                *self._grid.storagevalue_filling[this_type_filling][filling_col]
                *self._grid.storagevalue_time[this_type_time][timestep])
            self.concretemodel.genCost[i] = storagevalue
            if i in self._idx_generatorsWithPumping:
                deadband = self._grid.generator.pump_deadband[i]
                self.concretemodel.pumpCost[i] = storagevalue - deadband

        # 3b. flexible load (storage value)
        for i in self._idx_consumersWithFlexLoad:
            this_type_filling = self._grid.consumer['flex_storval_filling'][i]
            this_type_time = self._grid.consumer['flex_storval_time'][i]
            # Compute storage capacity in Mwh (from value in hours)
            storagecapacity_flexload = (
                self._grid.consumer['flex_storage'][i]      # h
                * self._grid.consumer['flex_fraction'][i]   #
                * self._grid.consumer['demand_avg'][i])     # MW
            fillinglevel = (
                self._storage_flexload[i] / storagecapacity_flexload  )
            filling_col = int(round(fillinglevel*100))
            if fillinglevel > 1:
                storagevalue_flex = -const.flexload_outside_cost
            elif fillinglevel < 0:
                storagevalue_flex = const.flexload_outside_cost
            else:
                storagevalue_flex = (
                    self._grid.consumer.flex_basevalue[i]
                    *self._grid.storagevalue_filling[this_type_filling][filling_col]
                    *self._grid.storagevalue_time[this_type_time][timestep])
            self.concretemodel.flexLoadCost[i] = storagevalue_flex


        return


    def _updatePowerLosses(self,aclossmultiplier=1,dclossmultiplier=1):
        '''Compute power losses from OPF solution and update parameters'''
        if self._lossmethod==0:
            pass
        elif self._lossmethod==1:            
            # Use constant loss parameters
            # If loss parameters should change, they need to be declared
            # mutable=True
            pass
        elif self._lossmethod==2:
            # Losses from previous timestep added as load
            for b in self.concretemodel.BRANCH_AC:
#                # r and x are given in pu; theta
#                loss_pu = r * ((theta_to-theta_from)*const.baseAngle/x)**2
#                # convert from p.u. to physical unit
#                lossMVA = loss_pu*const.baseMVA
                #TODO: simpler (check and replace):
                r = self._grid.branch.loc[b,'resistance']
                lossMVA = r * self.concretemodel.varAcBranchFlow[b]**2/const.baseMVA
                # A multiplication factor to account for reactive current losses
                # (or more precicely, to get similar results as Giacomo in 
                # the SmartNet project)
                lossMVA = lossMVA * aclossmultiplier
                self.concretemodel.branchAcPowerLoss[b] = lossMVA           
            for b in self.concretemodel.BRANCH_DC:
                #TODO: Test this before adding
                r_pu = self._grid.dcbranch.loc[b,'resistance']
                p_pu = self.concretemodel.varDcBranchFlow[b]/const.baseMVA
                loss_pu = r_pu * p_pu**2
                lossMVA = loss_pu * const.baseMVA * dclossmultiplier
                self.concretemodel.branchDcPowerLoss[b] = lossMVA
                #self.concretemodel.branchDcPowerLoss[b] = 0.0
        else:
            raise Exception("Loss method={} is not implemented"
                            .format(self._lossmethod))
        



    def _storeResultsAndUpdateStorage(self,timestep,results):
        """Store timestep results in local arrays, and update storage"""


        # 1. Update generator storage:
        inflow_profile_refs = self._grid.generator['inflow_ref']
        inflow_factor = self._grid.generator['inflow_fac']
        capacity= self._grid.generator['pmax']
        pumpedIn = np.zeros(len(capacity))
        energyIn = np.zeros(len(capacity))
        energyOut = np.zeros(len(capacity))
        for i in self.concretemodel.GEN:
            genInflow = (capacity[i] * inflow_factor[i]
        			 * self._grid.profiles[inflow_profile_refs[i]][timestep] )
            energyIn[i] = genInflow*self.timeDelta
            energyOut[i] = (self.concretemodel.varGeneration[i].value
                            *self.timeDelta)

        for i in self._idx_generatorsWithPumping:
            Ppump = self.concretemodel.varPump[i].value
            pumpedIn[i] = (Ppump*self._grid.generator['pump_efficiency'][i]
                            *self.timeDelta)
        energyStorable = (self._storage + energyIn + pumpedIn - energyOut)
        storagecapacity = self._grid.generator['storage_cap']
        #self._storage[i] = min(storagecapacity,energyStorable)
        self._storage = np.vstack((storagecapacity,energyStorable)).min(axis=0)
        self._energyspilled = energyStorable-self._storage

        # 2. Update flexible load storage
        for i in self._idx_consumersWithFlexLoad:
            energyIn_flexload = (self.concretemodel.varFlexLoad[i].value
                                 *self.timeDelta)
            energyOut_flexload = ( self._grid.consumer['flex_fraction'][i]
                                    * self._grid.consumer['demand_avg'][i]
                                    * self.timeDelta )
            self._storage_flexload[i] += energyIn_flexload-energyOut_flexload


        # 3. Collect variable values from optimisation result
        F = self.concretemodel.OBJ()
        Pgen = [self.concretemodel.varGeneration[i].value
                for i in self.concretemodel.GEN]
        Ppump = [self.concretemodel.varPump[i].value
                for i in self.concretemodel.GEN_PUMP]
        Pflexload = [self.concretemodel.varFlexLoad[i].value
                for i in self.concretemodel.LOAD_FLEX]
        Pb = [self.concretemodel.varAcBranchFlow[i].value
                for i in self.concretemodel.BRANCH_AC]
        Pdc = [self.concretemodel.varDcBranchFlow[i].value
                for i in self.concretemodel.BRANCH_DC]
        theta = [self.concretemodel.varVoltageAngle[i].value*const.baseAngle
                for i in self.concretemodel.NODE]
        #load shedding is aggregated to nodes (due to old code)
        Ploadshed = pd.Series(index=self._grid.node.id,
                                 data=[0]*len(self._grid.node.id),
                                 dtype=float)
        for j in self.concretemodel.LOAD:
            node = self._grid.consumer['node'][j]
            Ploadshed[node] += self.concretemodel.varLoadShed[j].value

        # 4 Collect dual values
        # 4a. branch capacity sensitivity (whether pos or neg flow)
        senseB = []
        for j in self._idx_branchesWithConstraints:
        #for j in self.concretemodel.BRANCH_AC:
            c = self.concretemodel.cMaxFlowAc[j]
            senseB.append(-abs(self.concretemodel.dual[c]/const.baseMVA ))
        senseDcB = []
        for j in self.concretemodel.BRANCH_DC:
            c = self.concretemodel.cMaxFlowDc[j]
            senseDcB.append(-abs(self.concretemodel.dual[c]/const.baseMVA ))

        # 4b. node demand sensitivity (energy balance)
        # TODO: Without abs(...) the value jumps between pos and neg. Why?
        senseN = []
        for j in self.concretemodel.NODE:
            c = self.concretemodel.cPowerbalance[j]
            senseN.append(abs(self.concretemodel.dual[c]/const.baseMVA ))

        # consider spilled energy only for generators with storage<infinity
        #energyspilled = zeros(energyStorable.shape)
        #indx = self._grid.getIdxGeneratorsWithNonzeroInflow()
        #energyspilled[indx] = energyStorable[indx]-self._storage[indx]
        energyspilled = self._energyspilled
        storagelevel = self._storage[self._idx_generatorsWithStorage]
        storageprice = [self.concretemodel.genCost[i].value
                            for i in self._idx_generatorsWithStorage]
        flexload_storagelevel = self._storage_flexload[self._idx_consumersWithFlexLoad]
        flexload_marginalprice = [self.concretemodel.flexLoadCost[i].value
                                    for i in self._idx_consumersWithFlexLoad]

        # TODO: Only keep track of inflow spilled for generators with
        # nonzero inflow

        # Extract power losses
        if self._lossmethod==0:
            acPowerLoss = [0]*len(self.concretemodel.BRANCH_AC)
            dcPowerLoss = [0]*len(self.concretemodel.BRANCH_DC)
        elif self._lossmethod==1:
            acPowerLoss = [self.concretemodel.varLossAc12[b].value
                           +self.concretemodel.varLossAc21[b].value
                            for b in self.concretemodel.BRANCH_AC]
            dcPowerLoss = [self.concretemodel.varLossDc12[b].value
                           +self.concretemodel.varLossDc21[b].value
                            for b in self.concretemodel.BRANCH_DC]
#            acPowerLoss = [pyo.value(self.concretemodel.varAcBranchFlow12[b]
#                            *self.concretemodel.lossAcA[b]
#                                + self.concretemodel.lossAcB[b]
#                           +self.concretemodel.varAcBranchFlow21[b]
#                            *self.concretemodel.lossAcA[b]
#                                + self.concretemodel.lossAcB[b])
#                            for b in self.concretemodel.BRANCH_AC]
#            dcPowerLoss = [pyo.value(self.concretemodel.varDcBranchFlow12[b]
#                            *self.concretemodel.lossDcA[b] 
#                                + self.concretemodel.lossDcB[b]
#                           +self.concretemodel.varDcBranchFlow21[b]
#                            *self.concretemodel.lossDcA[b] 
#                                + self.concretemodel.lossDcB[b])
#                            for b in self.concretemodel.BRANCH_DC]
        elif self._lossmethod==2:
            acPowerLoss = list(self.concretemodel.branchAcPowerLoss.extract_values().values())
            dcPowerLoss = list(self.concretemodel.branchDcPowerLoss.extract_values().values())
        else:
            raise Exception("Lossmethod must be 0,1 or 2")

        results.addResultsFromTimestep(
            timestep = self._grid.timerange[0]+timestep,
            objective_function = F,
            generator_power = Pgen,
            generator_pumped = Ppump,
            branch_power = Pb,
            dcbranch_power = Pdc,
            node_angle = theta,
            sensitivity_branch_capacity = senseB,
            sensitivity_dcbranch_capacity = senseDcB,
            sensitivity_node_power = senseN,
            storage = storagelevel.tolist(),
            inflow_spilled = energyspilled.tolist(),
            loadshed_power = Ploadshed.tolist(),
            marginalprice = storageprice,
            flexload_power = Pflexload,
            flexload_storage = flexload_storagelevel.tolist(),
            flexload_storagevalue = flexload_marginalprice,
            branch_ac_losses = acPowerLoss,
            branch_dc_losses = dcPowerLoss
            )

        return


    def solve(self,results,solver='cbc',solver_path=None,warmstart=False,
              savefiles=False,aclossmultiplier=1,
              dclossmultiplier=1,logfile="lpsolver_log.txt"):
        '''
        Solve LP problem for each time step in the time range

        Parameters
        ----------
        results : Results
            PowerGAMA Results object reference
        solver : string (optional)
            name of solver to use ("cbc" or "gurobi"). Gurobi uses python
            interface, whilst CBC uses command line executable
        solver_path :string (optional, only relevant for cbc)
            path for solver executable
        warmstart : Boolean
            Use warmstart option (only some solvers, e.g. gurobi)
        savefiles : Boolean
            Save Pyomo model file and LP problem MPS file for each timestep
            This may be useful for debugging.
        aclossmultiplier : float
            Multiplier factor to scale computed AC losses, used with method 1
        dclossmultiplier : float
            Multiplier factor to scale computed DC losses, used with method 1
        logfile : string
            Name of log file for LP solver. Will keep only last iteration

        Returns
        -------
        results : Results
            PowerGAMA Results object reference
        '''

        # Initalise solver, and check it is available
        if solver=="gurobi":
            opt = pyo.SolverFactory('gurobi',solver_io='python')
            print(":) Using direct python interface to solver")
        elif solver=="gurobi_persistent":
            opt = pyo.SolverFactory('gurobi_persistent')
            print(":) Using persistent (in-memory) python interface to solver")
            print(" => remember to notify solver of model changes!")
            print("    https://pyomo.readthedocs.io/en/latest/solvers/persistent_solvers.html")
        else:
            solver_io = None
            #if solver=="cbc":
            # NL requres CBC with ampl interface built in
            #    solver_io="nl" 
            opt = pyo.SolverFactory(solver,executable=solver_path,
                                    solver_io=solver_io) 
            if opt.available():
                print (":) Found solver here: {}".format(opt.executable()))
            else:
                print(":( Could not find solver {}. Returning."
                        .format(solver))
                raise Exception("Could not find LP solver {}"
                                .format(solver))

        #TODO: Code for persistent solver
        # https://pyomo.readthedocs.io/en/latest/solvers/persistent_solvers.html
        # to use pwersistent solvers, probably have to set instance at the 
        # start, and then modify it in each iteration rather than giving 
        # it as an argument to opt.solve:
        #    opt.set_instance(self.concretemodel)
        # and then use opt.solve()
        # To modify e.g. a constraint between solves, remove and add, e.g.:
        #    opt.remove_constraint(m.c)  
        #    del m.c  
        #    m.c = pe.Constraint(expr=m.y <= m.x)  
        #    opt.add_constraint(m.c) 
        # Variables can be updated without removing/adding
        #    m.x.setlb(1.0)  
        #    opt.update_var(m.x)

        #Enable access to dual values
        self.concretemodel.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

        if self._lossmethod==2:
            print("Computing losses in first timestep")
            self._updateLpProblem(timestep=0)
            res = opt.solve(self.concretemodel,
                        tee=False, #stream the solver output
                        keepfiles=False, #print the LP file for examination
                        symbolic_solver_labels=True) # use human readable names
                         
            #Now, power flow values are computed for the first timestep, and
            # power losses can be computed.
            
        print("Solving...")
        numTimesteps = len(self._grid.timerange)
        count = 0
        warmstart_now=False
        for timestep in range(numTimesteps):
            # update LP problem (inflow, storage, profiles)
            self._updateLpProblem(timestep)
            self._updatePowerLosses(aclossmultiplier,dclossmultiplier)

            # solve the LP problem
            if savefiles:
                #self.concretemodel.pprint('concretemodel_{}.txt'.format(timestep))
                self.concretemodel.write("LPproblem_{}.mps".format(timestep),
                                 io_options={'symbolic_solver_labels':True})
                #self.concretemodel.write("LPproblem_{}.nl".format(timestep))

            if warmstart and opt.warm_start_capable():
                #warmstart available (does not work with cbc)
                if count>0:
                    warmstart_now=warmstart
                count = count+1
                res = opt.solve(self.concretemodel,
                        tee=False, #stream the solver output
                        keepfiles=False, #print the LP file for examination
                        warmstart=warmstart_now,
                        symbolic_solver_labels=True, # use human readable names
                        logfile=logfile)
            elif not warmstart:
                #no warmstart option
                res = opt.solve(self.concretemodel,
                        tee=False, #stream the solver output
                        keepfiles=False, #print the LP file for examination
                        symbolic_solver_labels=True, # use human readable names
                        logfile=logfile)
            else:
                raise Exception("Solver ({}) is not capable of warm start"
                                    .format(opt.name))

            # Results loaded automatically, so this is not required
            # self.concretemodel.solutions.load_from(res)

            # store result for inspection if necessary
            self.solver_res = res

            #debugging:
            if False:
                print("Solver status = {}. Termination condition = {}"
                    .format(res.solver.status,
                            res.solver.termination_condition))

            if (res.solver.status != pyomo.opt.SolverStatus.ok):
                warnings.warn("Something went wrong with LP solver: {}"
                                .format(res.solver.status))
                raise Exception("Something went wrong with LP solver: {}"
                                .format(res.solver.status))
            elif (res.solver.termination_condition
                    == pyomo.opt.TerminationCondition.infeasible):
                warnings.warn("t={}: No feasible solution found."
                                .format(timestep))
                raise Exception("t={}: No feasible solution found."
                                .format(timestep))

            self._update_progress(timestep,numTimesteps)

            # store results and update storage levels
            self._storeResultsAndUpdateStorage(timestep,results)

        return results


    def _update_progress(self,n,maxn):
        if self._fancy_progressbar:
            barLength = 20
            progress = float(n+1)/maxn
            block = int(round(barLength*progress))
            text = "\rProgress: [{0}] {1} ({2}%)  " \
                .format( "="*block + " "*(barLength-block),
                        n, int(progress*100))
            sys.stdout.write(text)
            sys.stdout.flush()
        else:
            if int(100*(n+1)/maxn) > int(100*n/maxn):
                sys.stdout.write("%d%% "% (int(100*(n+1)/maxn)))
                sys.stdout.flush()

    def setProgressBar(self,value):
        '''Specify how to show simulation progress

        Parameters
        ----------
        value : string
            'fancy' or 'default'
        '''
        if value=='fancy':
            self._fancy_progressbar=True
        elif value=='default':
            self._fancy_progressbar=False
        else:
            raise Exception('Progress bar bust be either "default" or "fancy"')



# -*- coding: utf-8 -*-
'''
Module containing PowerGAMA GridData class and sub-classes

Grid data and time-dependent profiles
'''

import pandas as pd
import numpy
import scipy.sparse
import math                                 # Used in myround
import networkx as nx


##=============================================================================



class GridData(object):
    '''
    Class for grid data storage and import
    '''

    # Headers and default values in input files:
    # default=None: column _must_ be present in input file
    keys_powergama = {
        'node': {'id':None, 'area':None, 'lat':None,'lon':None},
        'branch': {'node_from':None,'node_to':None,
                   'reactance':None,'capacity':None,'resistance':0},
        'dcbranch': {'node_from':None, 'node_to':None, 'capacity':None,
                     'resistance':0},
        'generator': {'type':None,'desc':'', 'node':None,
                       'pmax':None,'pmin':None,'fuelcost':None,
                       'inflow_fac':None,'inflow_ref':None,
                       'storage_cap':0,'storage_price':0,'storage_ini':0,
                       'storval_filling_ref':'','storval_time_ref':'',
                       'pump_cap':0,'pump_efficiency':0,'pump_deadband':0},
        'consumer' : {'node':None, 'demand_avg':None,'demand_ref':None,
                      'flex_fraction':0, 'flex_on_off':0, 'flex_basevalue':0,
                      'flex_storage':0, 'flex_storval_filling':'',
                      'flex_storval_time':'', 'flex_storagelevel_init':0.5}
        }

    # Required fields for investment analysis input data
    # Default value = -1 means that it should be computed by the program
    keys_sipdata = {
        'node': {'id':None,'area':None, 'lat':None, 'lon':None,
                 'offshore':None,'type':None,
                 'existing':None,'cost_scaling':None},
        'branch': {'node_from':None,'node_to':None,'capacity':None,
                   'capacity2':0,'reactance':0,
                   'expand':None,'expand2':None, 'max_newCap':-1,'distance':-1,
                   'cost_scaling':None,'type':None},
        'dcbranch' :{'node_from':'', 'node_to':'', 'capacity':0,
                     'resistance':0},
        'generator': {'type':None,'node':None,'desc':'','pmax':None,'pmax2':0,
                      'pmin':None,
                      'expand':None,'expand2':None,'p_maxNew':-1, 'cost_scaling':1,
                      'fuelcost':None,'fuelcost_ref':None,'pavg':0,
                      'inflow_fac':None,'inflow_ref':None},
        'consumer': {'node':None, 'demand_avg':None,'emission_cap':-1,
                     'demand_ref':None}
        }


    def __init__(self):
        '''
        Create GridData object with data and methods for import and
        processing of PowerGAMA grid data
        '''
        self.node = None
        self.branch = None
        self.dcbranch = None
        self.generator = None
        self.consumer = None
        #self.inflowProfiles = None
        #self.demandProfiles = None
        self.profiles = None
        self.storagevalue_filling = None
        self.storagevalue_time = None
        self.timeDelta = None
        self.timerange = None

        self.CSV_SEPARATOR = None #automatically detect

    def _myround(x, base=1,method='round'):
        '''Round to nearest multiple of base'''
        if method=='round':
            return int(base * round(float(x)/base))
        elif method=='floor':
            return int(base * math.floor(float(x)/base))
        elif method=='ceil':
            return int(base * math.ceil(float(x)/base))
        else:
            raise Exception("Rounding error")


    def readGridData(self,nodes,ac_branches,dc_branches,generators,consumers,
                     remove_extra_columns=False):
        '''Read grid data from files into data variables'''

        self.node = pd.read_csv(nodes,
                                dtype={'id':str, 'area':str})
        self.branch = pd.read_csv(ac_branches,
                                  dtype={'node_from':str,'node_to':str,
                                         'capacity':float})
        if not dc_branches is None:
            self.dcbranch = pd.read_csv(dc_branches,
                                        dtype={'node_from':str,'node_to':str,
                                               'capacity':float})
        else:
            self.dcbranch = pd.DataFrame(
                columns=self.keys_powergama['dcbranch'].keys())
        self.generator = pd.read_csv(generators,
                                     dtype={'node':str,'type':str})
        self.consumer = pd.read_csv(consumers,
                                    dtype={'node':str})

        self._checkGridDataFields(self.keys_powergama)


        #TODO: fix difference in node index between powergama and powergim
        # numerical index is needed in method computePowerFlowMatrices
        # (and maybe elsewhere). This does _not_ work with powergama:
        #self.node.set_index('id',inplace=True,append=False)
        #self.node['id']=self.node.index
        self._checkGridData()
        self._addDefaultColumns(keys=self.keys_powergama,
                                remove_extra_columns=remove_extra_columns)
        self._fillEmptyCells(keys=self.keys_powergama)

    def readSipData(self,nodes,branches,generators,consumers):
        '''Read grid data for investment analysis from files (PowerGIM)

        This is used with the grid investment module (PowerGIM)

        time-series data may be used for
        consumer demand
        generator inflow (e.g. solar and wind)
        generator fuelcost (e.g. one generator with fuelcost = power price)
        '''
        self.node = pd.read_csv(nodes,
                                #usecols=self.keys_sipdata['node'],
                                dtype={'id':str, 'area':str})
        #TODO use integer range index instead of id string, cf powergama
        self.node.set_index('id',inplace=True)
        self.node['id']=self.node.index
        self.node.index.name = 'index'
        self.branch = pd.read_csv(branches,
                                  #usecols=self.keys_sipdata['branch'],
                                  dtype={'node_from':str,'node_to':str})
        # dcbranch variable only needed for powergama.plotMapGrid
        self.dcbranch = pd.DataFrame()
        self.generator = pd.read_csv(generators,
                                     #usecols=self.keys_sipdata['generator'],
                                     dtype={'node':str,'type':str})
        self.consumer = pd.read_csv(consumers,
                                    #usecols=self.keys_sipdata['consumer'],
                                    dtype={'node':str})


        self._checkGridDataFields(self.keys_sipdata)
        self._addDefaultColumns(keys=self.keys_sipdata)
        self._fillEmptyCells(keys=self.keys_sipdata)
        self._checkGridData()



    def _fillEmptyCells(self,keys):
        '''Use default data where none is given'''
        #generators:
        for col,val in keys['generator'].items():
            if val != None:
                self.generator[col] = self.generator[col].fillna(
                    keys['generator'][col])
        #consumers:
        for col,val in keys['consumer'].items():
            if val != None:
                self.consumer[col] = self.consumer[col].fillna(
                    keys['consumer'][col])

        #branches:
        for col,val in keys['branch'].items():
            if val != None:
                self.branch[col] = self.branch[col].fillna(
                    keys['branch'][col])


    def _addDefaultColumns(self,keys,remove_extra_columns=False):
        '''insert optional columns with default values when none
        are provided in input files'''
        for k in keys['generator']:
            if k not in self.generator.keys():
                self.generator[k] = keys['generator'][k]
        for k in keys['consumer']:
            if k not in self.consumer.keys():
                self.consumer[k] = keys['consumer'][k]
        for k in keys['branch']:
            if k not in self.branch.keys():
                self.branch[k] = keys['branch'][k]
        for k in keys['dcbranch']:
            if k not in self.dcbranch.keys():
                self.dcbranch[k] = keys['dcbranch'][k]

        # Discard extra columns (comments etc)
        if remove_extra_columns:
            self.node = self.node[list(keys['node'].keys())]
            self.branch = self.branch[list(keys['branch'].keys())]
            self.dcbranch = self.dcbranch[list(keys['dcbranch'].keys())]
            self.generator = self.generator[list(keys['generator'].keys())]
            self.consumer = self.consumer[list(keys['consumer'].keys())]


    def _checkGridDataFields(self,keys):
        '''check if all required columns are present
        (ie. all columns with no default value)'''
        for k,v in keys['node'].items():
            if v==None and not k in self.node:
                raise Exception("Node input file must contain %s" %k)
        for k,v in keys['branch'].items():
            if v==None and not k in self.branch:
                raise Exception("Branch input file must contain %s" %k)
        for k,v in keys['dcbranch'].items():
            if v==None and not k in self.dcbranch:
                raise Exception("DC branch input file must contain %s" %k)
        for k,v in keys['generator'].items():
            if v==None and not k in self.generator:
                raise Exception("Generator input file must contain %s" %k)
        for k,v in keys['consumer'].items():
            if v==None and not k in self.consumer:
                raise Exception("Consumer input file must contain %s" %k)



    def _checkGridData(self):
        '''Check consistency of grid data'''

        #generator nodes
        for g in self.generator['node']:
            if not g in self.node['id'].values:
                raise Exception("Generator node does not exist: '%s'" %g)
        #consumer nodes
        for c in self.consumer['node']:
            if not c in self.node['id'].values:
                raise Exception("Consumer node does not exist: '%s'" %c)

        #branch nodes
        for c in self.branch['node_from']:
            if not c in self.node['id'].values:
                raise Exception("Branch from node does not exist: '%s'" %c)
        for c in self.branch['node_to']:
            if not c in self.node['id'].values:
                raise Exception("Branch to node does not exist: '%s'" %c)

        #dcbranch nodes
        for c in self.dcbranch['node_from']:
            if not c in self.node['id'].values:
                raise Exception("DC Branch from node does not exist: '%s'" %c)
        for c in self.dcbranch['node_to']:
            if not c in self.node['id'].values:
                raise Exception("DC Branch to node does not exist: '%s'" %c)


    def _readProfileFromFile(self,filename,timerange):
        profiles = pd.read_csv(filename,sep=self.CSV_SEPARATOR,engine='python')
        profiles = profiles.loc[timerange]
        profiles.index = range(len(timerange))
        return profiles

    def _readStoragevaluesFromFile(self,filename):
        profiles = pd.read_csv(filename,sep=self.CSV_SEPARATOR,engine='python')
        return profiles


    def readProfileData(self,filename,timerange,
                        storagevalue_filling=None,
                        storagevalue_time=None,
                        timedelta=1.0):
        """Read profile (timeseries) into numpy arrays"""

        #self.inflowProfiles = self._readProfileFromFile(inflow,timerange)
        #self.demandProfiles = self._readProfileFromFile(demand,timerange)
        self.profiles = self._readProfileFromFile(filename,timerange)
        self.timerange = timerange
        self.timeDelta = timedelta

        '''
        Storage values have both time dependence and filling level dependence

       The dependence is on filling level (0-100%), is given as an array
        with 101 elements
        '''
        if not storagevalue_filling is None:
            self.storagevalue_time = self._readProfileFromFile(
                storagevalue_time,timerange)
            self.storagevalue_filling = self._readStoragevaluesFromFile(
                storagevalue_filling)
        return



    def writeGridDataToFiles(self,prefix):
        '''
        Save data to new input files
        '''

        file_nodes = prefix+"node.csv"
        file_branches = prefix+"branch.csv"
        file_consumers = prefix+"consumer.csv"
        file_generators = prefix+"generator.csv"
        file_dcbranch = prefix+"dcbranch.csv"

        sep = self.CSV_SEPARATOR
        if sep is None:
            sep=','


        self.node.to_csv(file_nodes,sep=sep)
        self.branch.to_csv(file_branches,sep=sep)
        self.consumer.to_csv(file_consumers,sep=sep)
        self.generator.to_csv(file_generators,sep=sep)
        self.dcbranch.to_csv(file_dcbranch,sep=sep)
        return


    def numConsumers(self):
        return self.consumer.shape[0]

    def numGenerators(self):
        return self.generator.shape[0]

    def numNodes(self):
        return self.node.shape[0]

    def numBranches(self):
        return self.branch.shape[0]

    def numDcBranches(self):
        if not self.dcbranch is None:
            return self.dcbranch.shape[0]
        else:
            return 0

    def branchFromNodeIdx(self):
        """get node indices for branch FROM node"""
        #return [self.node[self.node['id']==self.branch['node_from'][k]]
        #        .index.tolist()[0] for k in range(self.numBranches())]
        return [self.node[self.node['id']==b['node_from']].index.tolist()[0]
                for i,b in self.branch.iterrows()]

    def branchToNodeIdx(self):
        """get node indices for branch TO node"""
        return [self.node[self.node['id']==self.branch['node_to'][k]]
                .index.tolist()[0] for k in self.branch.index.tolist()]

    def dcBranchFromNodeIdx(self):
        """get node indices for dc branch FROM node"""
        return [self.node[self.node['id']==self.dcbranch['node_from'][k]]
                .index.tolist()[0] for k in self.dcbranch.index.tolist()]

    def dcBranchToNodeIdx(self):
        """get node indices for dc branch TO node"""
        return [self.node[self.node['id']==self.dcbranch['node_to'][k]]
                .index.tolist()[0] for k in self.dcbranch.index.tolist()]


    def getGeneratorsAtNode(self,nodeIdx):
        """Indices of all generators attached to a particular node"""
        #indices = [i for i, x in enumerate(self.generator['node'])
        #            if x == self.node['id'][nodeIdx]]
        indices = self.generator['node'][self.generator.loc[:,'node']
                        ==self.node['id'][nodeIdx]].index.tolist()
        return indices

    def getGeneratorsWithPumpAtNode(self,nodeIdx):
        """Indices of all pumps attached to a particular node"""
        #indices = [i for i, x in enumerate(self.generator['node'])
        #            if x == self.node['id'][nodeIdx]
        #            and self.generator['pump_cap'][i]>0]
        indices = self.generator['node'][
                    (self.generator.loc[:,'node']==self.node['id'][nodeIdx])
                    & (self.generator.loc[:,'pump_cap']>0)].index.tolist()
        return indices

    def getLoadsAtNode(self,nodeIdx):
        """Indices of all loads (consumers) attached to a particular node"""
        #indices = [i for i, x in enumerate(self.consumer['node'])
        #            if x == self.node['id'][nodeIdx]]:
        #25 times faster:
        indices = self.consumer['node'][self.consumer.loc[:,'node']
                                ==self.node['id'][nodeIdx]].index.tolist()
        return indices

    def getLoadsFlexibleAtNode(self,nodeIdx):
        """Indices of all flexible nodes attached to a particular node"""
        #indices = [i for i, x in enumerate(self.consumer['node'])
        #            if x == self.node['id'][nodeIdx]
        #            and self.consumer['flex_fraction'][i]>0
        #            and self.consumer['demand_avg'][i]>0]
        #faster:
        indices = self.consumer['node'][
            (self.consumer.loc[:,'node']==self.node['id'][nodeIdx])
            & (self.consumer.loc[:,'flex_fraction'] > 0)
            & (self.consumer.loc[:,'demand_avg'] > 0)].index.tolist()
        return indices

    def getIdxConsumersWithFlexibleLoad(self):
        """Indices of all consumers with flexible load"""
        idx = [i for i,v in enumerate(self.consumer['flex_fraction'])
            if v>0 and v<numpy.inf and self.consumer['demand_avg'][i]>0]
        return idx

    def getFlexibleLoadStorageCapacity(self,consumer_indx):
        ''' flexible load storage capacity in MWh'''
        cap = (self.consumer['demand_avg'][consumer_indx]
                * self.consumer['flex_fraction'][consumer_indx]
                * self.consumer['flex_storage'][consumer_indx] )
        return cap


    def getDcBranchesAtNode(self,nodeIdx,direction):
        """Indices of all DC branches attached to a particular node"""
        if direction=='from':
            indices = [i for i, x in enumerate(self.dcbranch['node_from'])
            if x == self.node['id'][nodeIdx]]
        elif direction=='to':
            indices = [i for i, x in enumerate(self.dcbranch['node_to'])
            if x == self.node['id'][nodeIdx]]
        else:
            raise Exception("Unknown direction in GridData.getDcBranchesAtNode")
        return indices


    def getDcBranches(self):
        '''
        Returns a list with DC branches in the format
        [index,from area,to area]
        '''
        hvdcBranches = []
        for idx in self.dcbranch.index.tolist():
            fromNodeIdx = self.node['id'].index(self.dcbranch['node_from'][idx])
            toNodeIdx = self.node.name.index(self.dcbranch['node_to'][idx])
            areaFrom = self.node['area'][fromNodeIdx]
            areaTo = self.node['area'][toNodeIdx]
            hvdcBranches.append([idx,areaFrom,areaTo])
        return hvdcBranches


    def getIdxNodesWithLoad(self):
        """Indices of nodes that have load (consumer) attached to them"""
        # Get index of node associated with all consumer
        #indices = numpy.asarray(self.consumer.nodeIdx(self.node))
        ## Return indices only once (unique values)
        #indices = numpy.unique(indices)
        loadnodes = self.node[self.node['id'].isin(self.consumer['node'])]
        indices = numpy.asarray(loadnodes.index)
        return indices


    def getIdxGeneratorsWithStorage(self):
        """Indices of all generators with nonzero and non-infinite storage"""
        idx = [i for i,v in enumerate(self.generator['storage_cap'])
            if v>0 and v<numpy.inf]
        return idx

    def getIdxGeneratorsWithNonzeroInflow(self):
        """Indices of all generators with nonzero inflow"""
        idx = [i for i,v in enumerate(self.generator['inflow_fac'])
            if v>0]
        return idx

    def getIdxGeneratorsWithPumping(self):
        """Indices of all generators with pumping capacity"""
        idx = [i for i,v in enumerate(self.generator['pump_cap'])
            if v>0 and v<numpy.inf]
        return idx

    def getIdxBranchesWithFlowConstraints(self):
        '''Indices of branches with less than infinite branch capacity'''
        idx = [i for i,v in enumerate(self.branch['capacity']) if v<numpy.inf]
        return idx

    def getIdxDcBranchesWithFlowConstraints(self):
        '''Indices of DC branches with less than infinite branch capacity'''
        if self.dcbranch is None:
            idx = []
        else:
            idx = [i for i,v in enumerate(self.dcbranch['capacity'])
                        if v<numpy.inf]
        return idx


    def getIdxBranchesWithLength(self):
        '''Indices of branches with specified length'''
        idx = [i for i,v in enumerate(self.branch['length']) if not math.isnan(v)]
        return idx

    def _susceptancePu(self,baseOhm=1):
        '''If impedance is already given in pu, baseOhm should be 1
        If not, well... baseOhm depends on the voltage level, so need to know
        the nominal voltage at the bus to convert from ohm to pu.
        '''
        if (self.branch['reactance']==0).any():
            raise Exception("Zero branch reactance is not allowed.")
        return [-1/self.branch['reactance'][i]*baseOhm 
                for i in self.branch.index.tolist()]


    def compute_power_flow_matrices(self, base_Z=1):
        """
        Compute and return dc power flow matrices B' and DA
        Parameters
        ==========
        base_Z : float (impedance should already be in pu.)
                base value for impedance
        Returns
        =======
        (coeff_B, coeff_DA) : scipy.sparse matrices
        """

        #df_branch = pd.DataFrame.from_dict(branches, orient="index")
        branches = self.branch
        susceptance = 1 / branches["reactance"] * base_Z
        nodes = list(self.node["id"])
        #node_ids = nodes
        edge_ids = []
        edges = []
        #for br_id, branch in branches.items():
        for br_id, branch in branches.iterrows():
            b = susceptance[br_id]
            edges.append((branch["node_from"], branch["node_to"], br_id, {"i": br_id, "b": b}))
            edge_ids.append(br_id)

        # MultiDiGraph to allow parallel lines
        G = nx.MultiDiGraph()
        G.add_nodes_from(nodes)
        G.add_edges_from(edges)
        A_incidence_matrix = -nx.incidence_matrix(G, oriented=True, nodelist=nodes, edgelist=edges).T       
        A_incidence_matrix = scipy.sparse.csc_matrix(A_incidence_matrix)

        # Diagonal matrix
        D = scipy.sparse.diags(-susceptance, offsets=0)
        # Element-wise multiplication:
        DA = D * A_incidence_matrix

        # Bf constructed from incidence matrix with branch susceptance
        # used as weight (this is quite fast)
        Bf = -nx.incidence_matrix(G, oriented=True, nodelist=nodes, edgelist=edges, weight="b").T
        # See note above for this conversion
        Bf = scipy.sparse.csc_matrix(Bf)
        # Element-wise multiplication:
        Bbus = A_incidence_matrix.T * Bf
        return Bbus, DA
    

    def getAllAreas(self):
        '''Return list of areas included in the grid model'''
        areas = self.node['area']
        allareas = []
        for co in areas:
            if co not in allareas:
                allareas.append(co)
        return allareas

    def getAllGeneratorTypes(self,sort='fuelcost'):
        '''Return list of generator types included in the grid model'''
        gentypes = self.generator.type
        alltypes = []
        if sort==None:
            gentypes = self.generator.type
            alltypes = []
            for ge in gentypes:
                if ge not in alltypes:
                    alltypes.append(ge)
            return alltypes
        elif sort=='fuelcost':
            generators = self.getGeneratorsPerType()
            avg = {ge_k : numpy.mean(self.generator.fuelcost[ge_v])
                   for ge_k,ge_v in generators.items()}
            sorted_list = [k for k in sorted(avg, key=avg.get, reverse=False)]

#            gentypes = generators.keys()
#            fuelcosts = []
#            for ge in gentypes:
#                gen_this_type = generators[ge]
#                fuelcosts.append(numpy.mean([self.generator.fuelcost[i]
#                                         for i in gen_this_type]) )
#            sorted_list = [x for (y,x) in
#                           sorted(zip(fuelcosts,gentypes))]
            return sorted_list
        else:
            raise Exception("sort must be None or 'fuelcost'")


    def getConsumerAreas(self):
        """List of areas for each consumer"""
        areas = [self.node.area[self.node['id']==n].tolist()[0]
                    for n in self.consumer['node']]
        return areas

    def getGeneratorAreas(self):
        """List of areas for each generator"""
        areas = [self.node.area[self.node['id']==n].tolist()[0]
                    for n in self.generator.node]
        return areas

    def getConsumersPerArea(self):
        '''Returns dictionary with indices of loads within each area'''
        consumers = {}
        consumer_areas = self.getConsumerAreas()
        for idx_load in self.consumer.index.tolist():
            area_name = consumer_areas[idx_load]
            if area_name in consumers:
                consumers[area_name].append(idx_load)
            else:
                consumers[area_name] =  [idx_load]
        return consumers

    def getGeneratorsPerAreaAndType(self):
        '''Returns dictionary with indices of generators within each area'''
        generators = {}
        generator_areas = self.getGeneratorAreas()
        for idx_gen in self.generator.index.tolist():
            gtype = self.generator.type[idx_gen]
            area_name = generator_areas[idx_gen]
            if area_name in generators:
                if gtype in generators[area_name]:
                    generators[area_name][gtype].append(idx_gen)
                else:
                    generators[area_name][gtype] = [idx_gen]
            else:
                generators[area_name] = {gtype:[idx_gen]}
        return generators

    def getGeneratorsPerType(self):
        '''Returns dictionary with indices of generators per type'''
        generators = {}
        for idx_gen in self.generator.index.tolist():
            gtype = self.generator['type'][idx_gen]
            if gtype in generators:
                generators[gtype].append(idx_gen)
            else:
                generators[gtype] = [idx_gen]
        return generators


    def getGeneratorsWithPumpByArea(self):
        '''
        Returns dictionary with indices of generators with pumps within
        each area
        '''
        generators = {}
        for pumpIdx,gen in self.generator.iterrows():
            cap = gen['pump_cap']
            if cap>0 and cap<numpy.inf:
                nodeId = self.generator.loc[pumpIdx,'node']
                nodeIdx = self.node[self.node['id']==nodeId].index[0]
                areaName = self.node.loc[nodeIdx,'area']
                if areaName in generators:
                    generators[areaName].append(pumpIdx)
                else:
                    generators[areaName] = [pumpIdx]
        return generators

    def getBranchAreas(self):
        br_witharea = pd.DataFrame()
        br_witharea['area_from'] = self.branch.merge(right=self.node[['id','area']],
                            how='left',left_on='node_from',
                            right_on='id')['area']
        br_witharea['area_to'] = self.branch.merge(right=self.node[['id','area']],
                            how='left',left_on='node_to',
                            right_on='id')['area']
        return br_witharea

    def getDcBranchAreas(self):
        br_witharea = self.dcbranch
        br_witharea['area_from'] = self.dcbranch.merge(right=self.node[['id','area']],
                            how='left',left_on='node_from',
                            right_on='id')['area']
        br_witharea['area_to'] = self.dcbranch.merge(right=self.node[['id','area']],
                            how='left',left_on='node_to',
                            right_on='id')['area']
        return br_witharea

    def getInterAreaBranches(self,area_from=None,area_to=None,acdc='ac'):
        '''
        Get indices of branches from and/or to specified area(s)

        area_from = area from. Use None (default) to leave unspecifie
        area_to= area to. Use None (default) to leave unspecified
        acdc = 'ac' (default) for ac branches, 'dc' for dc branches
        '''

        if area_from is None and area_to is None:
            raise Exception("Either from area or to area (or both) has"
                            +"to be specified)")

        if acdc=='ac':
            br_witharea = self.getBranchAreas()
        elif acdc=='dc':
            br_witharea = self.getDcBranchAreas()

        if area_to is not None:
            mask_pos = ((br_witharea['area_to']==area_to)
                       & (br_witharea['area_from']!=area_to))
            mask_neg = ((br_witharea['area_from']==area_to)
                        & (br_witharea['area_to']!=area_to))
        if area_from is not None:
            mask_pos2 = ((br_witharea['area_from']==area_from)
                     & (br_witharea['area_to']!=area_from) )
            mask_neg2 = ((br_witharea['area_to']==area_from)
                        & (br_witharea['area_from']!=area_from))
            if area_to is not None:
                mask_pos = mask_pos & mask_pos2
                mask_neg = mask_neg & mask_neg2
            else:
                mask_pos = mask_pos2
                mask_neg = mask_neg2

        branches_pos = list(br_witharea.index[mask_pos])
        branches_neg = list(br_witharea.index[mask_neg])

#        # indices of from and to nodes of all branches:
#        if acdc=='ac':
#            br_from_nodes = self.branchFromNodeIdx()
#            br_to_nodes = self.branchToNodeIdx()
#        elif acdc=='dc':
#            br_from_nodes = self.dcBranchFromNodeIdx()
#            br_to_nodes = self.dcBranchToNodeIdx()
#        else:
#            raise Exception('Branch type must be "ac" or "dc"')
#
#
#        br_from_area = [self.node.area[i] for i in br_from_nodes]
#        br_to_area = [self.node.area[i] for i in br_to_nodes]
#
#        # indices of all inter-area branches (from area != to area)
#        br_is_interarea = [i for i in range(len(br_from_area))
#                                if br_from_area[i] != br_to_area[i]]
#
#        # branches connected to area_from
#        fromArea_branches_pos = [i for i in br_is_interarea
#                                 if br_from_area[i]==area_from]
#        fromArea_branches_neg = [i for i in br_is_interarea
#                                 if br_to_area[i]==area_from]
#
#        # branches connected to area_to
#        toArea_branches_pos = [i for i in br_is_interarea
#                                 if br_to_area[i]==area_to]
#        toArea_branches_neg = [i for i in br_is_interarea
#                                 if br_from_area[i]==area_to]
#
#        if area_from is None:
#            # Only to node has been specified
#            branches_pos = toArea_branches_pos
#            branches_neg = toArea_branches_neg
#        elif area_to is None:
#            # Only from node has been specified
#            branches_pos = fromArea_branches_pos
#            branches_neg = fromArea_branches_neg
#        else:
#            # Both to and from area has been specified
#            branches_pos = [b for b in fromArea_branches_pos
#                                    if b in toArea_branches_neg ]
#            branches_neg = [b for b in fromArea_branches_neg
#                                    if b in toArea_branches_pos ]
        return dict(branches_pos=branches_pos,
                    branches_neg=branches_neg)


    def branchDistances(self,R=6373.0):
        '''computes branch distance from node coordinates, resuls in km

        Uses haversine formula

        Parameters
        ----------
        R : radius of the Earth
        '''

        # approximate radius of earth in km
        n_from = self.branchFromNodeIdx()
        n_to = self.branchToNodeIdx()
        distance = []
        # get endpoint coordinates and convert to radians
        lats1 = self.node['lat'][n_from].apply(math.radians)
        lons1 = self.node['lon'][n_from].apply(math.radians)
        lats2 = self.node['lat'][n_to].apply(math.radians)
        lons2 = self.node['lon'][n_to].apply(math.radians)
        lats1.index=self.branch.index
        lons1.index=self.branch.index
        lats2.index=self.branch.index
        lons2.index=self.branch.index

        for b in self.branch.index:
            lat1=lats1[b]
            lon1=lons1[b]
            lat2=lats2[b]
            lon2=lons2[b]

            dlon = lon2 - lon1
            dlat = lat2 - lat1

            a = (math.sin(dlat/2)**2
                + math.cos(lat1) * math.cos(lat2) * math.sin(dlon/2)**2 )
            c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
            #atan2 better than asin: c = 2 * math.asin(math.sqrt(a))
            distance.append(R * c)
        return distance

    def spreadNodeCoordinates(self,radius=0.01,inplace=False):
        '''Spread nodes with identical coordinates in a small circle
        with radius r

        Parameters
        ----------
        radius : float
            radius in degrees for the size of the spread
        inplace : boolean
            if true, update GridData object

        Returns
        -------
        coords : array
            lat,lon pandas array for nodes
        '''
        coords = self.node[['lat','lon']]
        dupl_coords = pd.DataFrame()
        dupl_coords['cumcount'] = coords.groupby(['lat','lon']).cumcount()
        dupl_coords['count'] = coords.groupby(['lat','lon'])['lon'].transform('count')
        coords_new = coords.copy()
        for i,c in coords.iterrows():
            n_sum = dupl_coords.loc[i,'count']
            if n_sum >1:
                # there are more nodes with the same coordinates
                n = dupl_coords.loc[i,'cumcount']
                theta = 2*math.pi/n_sum
                coords_new.loc[i,'lat'] += radius*math.cos(n*theta)
                coords_new.loc[i,'lon'] += radius*math.sin(n*theta)
        if inplace:
            self.node[['lat','lon']] = coords_new
        return coords_new
